/**
 * @file		AppConfig.h
 * @brief		Main configuration file for the System and the Application components.
 * The AppConfig.h file is evaluated first, as it is loaded using the --preinclude switch.
 * All definitions required to configure the application, drivers, and system should be placed in this file.
 * 
 * Review the \ref APPConfig module as well.
 */

#ifndef APPCONFIG_H_
#define APPCONFIG_H_


#include <stdint.h>

/**
 * @defgroup 	APPConfig Application Configuration
 * 
 * @brief		This module presents ways of configuring an application.
 * 
 * The System API and Drivers provide a set of definitions that can be tuned in order to modify the configuration
 * of the drivers and the system. The user can also specify his own Defines that
 * are required for his application to work properly. All relevant details can
 * be found in the file \ref AppConfig.h
 * 
 * The Application Configuration module also allows to define the persisten application settings.
 */

 /**
 * @name Persistent application settings
 * 
 * The System API is taking care of persistent storage of application settings. The user defines his set of application
 * settings by modifying the definition of the \ref appconfig_t struct in the \ref AppConfig.h file.
 * @n@n
 * After the power on event of the device, application settings are loaded and made available in the global \ref AppConfiguration variable.
 * Any changes of the application settings should be performed on the \ref AppConfiguration variable and persisted by calling the
 * Configuration_Save() function.
 * 
 * Example 1: Accessing the Configuration:
 * @code
 * LOG("Threshold: %u", AppConfiguration.Threshold);
 * @endcode
 * 
 * Example 2: Updating the Configuration:
 * @code
 * AppConfiguration.Threshold = 120;
 * Configuration_Save();
 * @endcode
 * 
 * @{
 */

/**
 * @brief User defined set of application settings. Settings are stored in \ref AppConfiguration variable and persisted by the System API.
 * 
 * @remarks The appconfig_t struct must not be empty.
 */
typedef struct {
	uint8_t		Example;		///< an example entry in the application configuration
} appconfig_t;

/**
 * @ingroup APPConfig
 * @brief	Global variable holding user defined application settings
 */
appconfig_t AppConfiguration;


/** @} */


/**
 * @name	Application configuration
 * 
 * The following definitions can be modified in order to tune the behavior and the configuration of the Application.
 * 
 * @{
 */

/**
 * @ingroup	APPConfig
 * @{
 */

/**
 * @brief		Controls the Autonomous Base Station. Select the name of the define for more details.
 * 
 * The Autonomous Base Station will start downloading data when a Tracker reports more new count of fixes than the value defined here.
 */
#define THRESHOLD_NEW_FIX_COUNT				10

/**
 * @brief		Controls the Autonomous Base Station. Select the name of the define for more details.
 * 
 * The Autonomous Base Station will start downloading data when a Tracker reports more new count of log entries than the value defined here.
 */
#define THRESHOLD_NEW_LOG_COUNT				10

/**
 * @brief		Controls the Autonomous Base Station. Select the name of the define for more details.
 * 
 * The Autonomous Base Station will start abort further search for missing entries if it detects than the percentage of missing 
 * entries is less than the value specified here. Independent of this setting, it will always try to download new entries once
 * and it will always try to find missing entries once.
 */
#define THRESHOLD_MISSING_ENTRIES			5

/** @} */
/** @} */

/**
 * @name	System configuration
 * 
 * The following definitions can be modified in order to tune the behavior and the configuration of the System and Drivers.
 * If you aren't sure about the proper value of the Define you can just remove it from the \ref AppConfig.h file,
 * in such case, a default value predefined by the System/Driver developer will be used.
 * 
 * @{
 */

/**
 * @ingroup	APPConfig
 * @{
 */

#define COMMUNICATION_PROTOCOL_VERSION		11	///< The version of the communication protocol used between the "base" and the "tracker" application. Both versions need to be equal, the "base" will ignore the "tracker" otherwise.

#define UART1_RX_BUFFER_SIZE				512	///< Size of the receive buffer used to store NMEA data received from the GPS device. Note: one NMEA sentence contains of around 80 characters. 

/*
 * The RMC sentence contains date and time information received from the GPS system.
 * The frequency with which the RMC sentece is provicdd by the GPS receiver depends on the system position rate.
 * The number provided here specifies the requency relative to the system position rate.
 * 
 * Example:
 * 	Setting the value to 5 means, the RMC sentence will be delivered every 5th time. Depending on the system position rate, e.g. when
 * 	the system position rate is set to 1Hz, the RMC sentence will be delivered every 5 seconds
 * 
 *  Supported values for the RMC update rate: 0, 1, 2, 3, 4, 5
 */
#define		GPS_RMC_UPDATE_RATE		"5"

/**
 * @brief Logging Engine: Controlls the verbose debug output. Select the name of the define for more details.
 *
 * Many components of the System components and Drivers generate extensive debug output over the
 * serial port. The debug output feature can be disabled using this switch.
 * 
 * Printing text is expensive in terms of code size, execution time (printing text is slow) and as 
 * a consequence, in terms of power consumption and life time of the device.
 * 
 * Please consider this during the application development and thing about disabling the debug output
 * when deploying the device.
 * 
 * @remarks 	Remember to test the entire application after disabling this feature.
 * 				The execution time of some functions will be reduced. This can cause a change in behavior of your application.
 * 
 * The debug output feature can be used in your application.
 * 
 * Example:
 * @code
 * // current state of the algorithm
 * uint8_t currentState;
 * // ... algorithm executes, the state of the algorithm changes
 * //...
 * LOGDBG("Current State = %u", currentState); // prints the current state to the debug output
 * @endcode
 */
#define LOGGING_ENABLE_DEBUG_OUTPUT			1

/**
 * @brief Commands Engine: Controls the functionality of the "help" command. Select the name of the define for more details.
 *
 * Setting it to 1 causes the engine to print short help about every command
 * when the help command is executed
 * 
 * @remarks 	The description of a command requireds extra space when your application is compiled.
 * 				Set this switch to 0 if you are running low of program memory.
 */
#define COMMAND_HELP_ENABLED				1

/**
 * @brief Logging Engine: Controlls the log output. Select the name of the define for more details.
 *
 * Components of the System components and Drivers generate some log output using this feature. The log output 
 * feature can be disabled using this switch.
 * 
 * Printing text is expensive in terms of code size, execution time (printing text is slow) and as 
 * a consequence, in terms of power consumption and life time of the device.
 * 
 * Please consider this during the application development and limit the amout of printed text.
 * 
 * @remarks 	Remember to test the entire application after disabling this feature.
 * 				The execution time of some functions will be reduced. This can cause a change in behavior of your application.
 * 
 * The log output feature can be used in your application.
 *  
 * Example:
 * @code
 * LOG("Experiment: done"); // prints information using the log output
 * @endcode
 */
#define LOGGING_ENABLE_LOG_OUTPUT			1

/**
 * @brief Logging Engine: Controlls the radio log output. Select the name of the define for more details.
 *
 * You can use this feature to print text on remote devices. This can be useful to distribute
 * status information during the development phase. The message will be transmitted over the radio
 * as a broadcast. All devices that hear the message and have this feature enabled as well, will 
 * print the received message over the serial port.
 * 
 * Printing and transmitting text over radio is expensive in terms of code size, execution time (printing text is slow) and as 
 * a consequence, in terms of power consumption and life time of the device.
 * 
 * Please consider this during the application development and limit the usage of this feature.
 * 
 * @remarks 	Remember to test the entire application after disabling this feature.
 * 				The execution time of some functions will be reduced. This can cause a change in behavior of your application.
 * 
 * The log output feature can be used in your application.
 *  
 * Example:
 * @code
 * LOGREMOTE("Received format request.");
 * @endcode
 */
#define LOGGING_ENABLE_REMOTE_LOG_OUTPUT	1

/**
 * @brief		Controls the LED feature on the device.
 * 
 * You can enable LED usage in the software by setting this switch to 1. Setting
 * this switch to 0 will disable the usage of the LED by the software.
 * 
 * @remarks		The periodic "alive" blink signal that happens every 8 seconds will execute independent of this setting. 
 */
#define LED_ENABLED						1

/** @} */
/** @} */


#endif /*APPCONFIG_H_*/

#ifndef DOXYGEN_PUBLIC_DOC
/**
* Copyright (c) Microsoft Corporation. All rights reserved.
* Licensed under the Apache License, Version 2.0 (the "License"); you may not use these files except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
* Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.
 
* Copyright (c) 2012 Freie Universitaet Berlin, Tomasz Naumowicz. All rights reserved.
* 
* Redistribution and use in source and binary forms, with or without modification, are
* permitted provided that the following conditions are met:
* 
* - Redistributions of source code must retain the above copyright notice, this list of
* conditions and the following disclaimer.
* 
* - Redistributions in binary form must reproduce the above copyright notice, this list
* of conditions and the following disclaimer in the documentation and/or other materials
* provided with the distribution.
* 
* - Neither the name of the Freie Universitaet Berlin nor the
* names of its contributors may be used to endorse or promote products
* derived from this software without specific prior written permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Freie Universitaet Berlin BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#endif // DOXYGEN
