/**
 * @brief		User Application: Autonomous Base Station
 */

#include "Application.API.h"

#include "FSM_Base.h"

#include "Application.h"

// sets the application name and version. the __APP_* values are defined in the makefile.
APP(__APP_NAME, __APPVER_MAJOR, __APPVER_MINOR);

// used for the alive heartbeat generated by the base station.
int16_t _reportTimerId;

// this function is executed periodicaly and repots status of the base station
// it registers itself for execution using the Timers_Add function. 
void ReportBeingAlive(uint16_t notUsed) {
	_reportTimerId = Timers_Add(ReportBeingAlive, TICKS_5MINUTES, 0);

	if (0 != Network_GetChannel()) {
		// something must be going on...
		return;
	}
	
	// battery voltage check
	logentry_t le;
	LogEngine_PrepareLogEntry(&le);
	le.TypeId = Battery;
		
	char battAsString[BATTERY_READ_STRING_LENGTH];
	SET(GPS_PWR_OUT, GPS_PWR_PIN);
	Timers_Block(4096);
	
	le.Payload.Battery.Value = ADC_Devices_Battery_Read(NULL, battAsString);
	
	CLEAR(GPS_PWR_OUT, GPS_PWR_PIN);

	LogEngine_Save(&le);
	
	// remote print
	LOGREMOTEANDLOCAL("Battery: %s, LogSize: %lu", battAsString, LogEngine_GetLogSize());
}


void Application_Init() {
	{
		// initialize all drivers
		init_drivers_result_t init_result = MATAKI_InitDrivers(); 

		// process the init results and ignore GPS failures as GSP is not required for the base station anyway
		if (Init_Success != init_result) {
			if (GPS_Failure == init_result)
				LOGERR("GPS Init failed: this app is not using GPS, ignoring this error.");
			else
				MATAKI_CriticalLPM4(1);
		}
	}
	
	LIGHT_DISABLE;
	
	//log Power On Reset event
	logentry_t entry;
	LogEngine_PrepareLogEntry(&entry);
	entry.TypeId = POR;
	LogEngine_Save(&entry);

	LOGREMOTE("base online! (1)");
	LOGREMOTE("base online! (2)");
	LOGREMOTE("base online! (3)");
	
	// initialize state machines
	FSM_Base_Init();
	
	_reportTimerId = Timers_Add(ReportBeingAlive, TICKS_5SECONDS, 0);
	
	LOGATT("Use the 'abort' command to abort the auto base mode (60 seconds)!");
	
	printf("Current Configuration:\n\r"
			"\t1. Start downloading when new log entries count > %u or when new fix count > %u\r\n"
			"\t2. Stop downloading when less than %u percent of data is missing.\r\n",
			THRESHOLD_NEW_LOG_COUNT,
			THRESHOLD_NEW_FIX_COUNT,
			THRESHOLD_MISSING_ENTRIES);
}

COMMAND(abort, "Aborts the automatic base station mode", args) {
	// abort the state machines
	if (FSM_Base_Abort()) {
		// remove the heartbeat only if FSM abort was successful
		Timers_Remove(_reportTimerId);
	}
}

COMMAND(f_reset, "Formats and resets the device", args) {
	printf("format + reset\r\n");

	LOGREMOTE("Formatting and restarting the device... (1)");
	LOGREMOTE("Formatting and restarting the device... (2)");
	LOGREMOTE("Formatting and restarting the device... (3)");
	
	Flash_Erase(); 								// format (format just exits without waiting for success.
	Flash_WaitForDeviceWithVisualFeedback();
	
	// and reset the device
	Watchdog_DeviceReset();
}



#ifndef DOXYGEN_PUBLIC_DOC
/**
* Copyright (c) Microsoft Corporation. All rights reserved.
* Licensed under the Apache License, Version 2.0 (the "License"); you may not use these files except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
* Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.
 
* Copyright (c) 2012 Freie Universitaet Berlin, Tomasz Naumowicz. All rights reserved.
* 
* Redistribution and use in source and binary forms, with or without modification, are
* permitted provided that the following conditions are met:
* 
* - Redistributions of source code must retain the above copyright notice, this list of
* conditions and the following disclaimer.
* 
* - Redistributions in binary form must reproduce the above copyright notice, this list
* of conditions and the following disclaimer in the documentation and/or other materials
* provided with the distribution.
* 
* - Neither the name of the Freie Universitaet Berlin nor the
* names of its contributors may be used to endorse or promote products
* derived from this software without specific prior written permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Freie Universitaet Berlin BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#endif // DOXYGEN
