#include "Application.API.h"

#include "FSM.h"
#include "FSM_Base.h"

#include "Shared/SharedTypes.h"

// 1. Define all actions here.
//    This could happen in the header file, but you might like to keep those functions local.
//    This definition is required for the definition of the transition table

static uint8_t _pingtracker_remaining_calls;
static uint8_t _evaluatedownload_remaining_calls;

static struct {
	uint32_t		ReceivedNewDataCounter;		///< tells how much was downloaded when the 'request' was issued
	uint32_t		ReceivedMissingCounter;		///< tells how much was downloaded across all 'fillgaps' requests
	uint32_t		RequestedMissingCounter;	///< summary of missing entries when the first 'fillgaps' executed
	uint32_t		LocalSequenceNo;			///< The local sequence number at whitch the download was started. important for gaps filling
	uint32_t		RemoteLogSize;
	uint16_t		RemoteNewLogEntriesCount;
	uint8_t			TrackerNodeId;
	uint8_t			CommChannel;
} _tracker_dl_status;
	
void PingTracker();
void PerformDownloadNewData();
void PerformDownloadMissingData();
void EvaluateDownload();
void ResetTrackerAndGoDefChannel();
void GoDefChannel();
void NoAction() {}

void NotifyAutoBSRun();

// 2. Define states you are going to use in your state machine
enum FSM_Base_States {
	FSM_Waiting				= 0,
	FSM_Idle 				= 1,
	FSM_Searching			= 2,
	FSM_LoadingNew 			= 3,
	FSM_LoadingMissing 		= 4,
	FSM_Evaluating			= 5
};

// 3. Count your states
#define FSM_Base_States_Count	6

// 4. Define events used in your state machine
//    Leave 'Timeout' at the beginning
enum FSM_Base_Events {
	Timeout					= 0,
	GoodHeartbeat 			= 1,
	TrackerNotFound 		= 2,
	TrackerFound 			= 3,
	DownloadPerformed 		= 4,
	EvaluationRetryRequest 	= 5,
	EvaluationGoodResult 	= 6,
	EvaluationFailure 		= 7,
	ForcedDownload			= 8
};

// 5. Count your events
#define FSM_Base_Events_Count 	9

// 6. Define timeouts for your states
//    When no event happens within the timeout value, the state change will be executed (signalled as timeout)
//    0 stands for no timeout.
uint32_t fsm_base_timeouts[FSM_Base_States_Count] = {
		0,					// FSM_Waiting
		0,					// FSM_Idle 		
		TICKS_2SECONDS,		// FSM_Searching	
		TICKS_10SECONDS,	// FSM_LoadingNew 	
		0,					// FSM_LoadingMissing 		
		0					// FSM_Evaluating	
};

// 7. Define your state transitions.
//    Horiziontal:	current state
//    Vertical:		The event
//    Entry:		Ation to call and the next state.
 FSM_Entry fsm_base_transitions[FSM_Base_Events_Count][FSM_Base_States_Count] = {
/*									Waiting							Idle								searching									loading new												loading missing							evaluating		*/
/* Timeout					*/	{ 	{ NotifyAutoBSRun, FSM_Idle }, 	{ NULL, 0 }, 						{ PingTracker, FSM_Searching },				{ PerformDownloadMissingData, FSM_LoadingMissing },		{ NULL, 0 }, 							{ NULL, 0 } },				
/* GoodHeartbeat			*/ 	{ 	{ NULL, 0 }, 					{ NoAction, FSM_Searching },		{ NULL, 0 },	 							{ NULL, 0 },											{ NULL, 0 }, 							{ NULL, 0 } },		
/* TrackerNotFound			*/ 	{ 	{ NULL, 0 }, 					{ NULL, 0 }, 						{ GoDefChannel, FSM_Idle },					{ NULL, 0 },											{ NULL, 0 }, 							{ NULL, 0 } },
/* TrackerFound				*/ 	{ 	{ NULL, 0 }, 					{ NULL, 0 }, 						{ PerformDownloadNewData, FSM_LoadingNew },	{ NULL, 0 },											{ NULL, 0 }, 							{ NULL, 0 } },
/* DownloadPerformed		*/ 	{ 	{ NULL, 0 }, 					{ NULL, 0 }, 						{ NULL, 0 },	 							{ NULL, 0 },											{ EvaluateDownload, FSM_Evaluating }, 	{ NULL, 0 } },	
/* EvaluationRetryRequest	*/ 	{ 	{ NULL, 0 }, 					{ NULL, 0 }, 						{ NULL, 0 },	 							{ NULL, 0 },											{ NULL, 0 }, 							{ PerformDownloadMissingData, FSM_LoadingMissing } },		
/* EvaluationGoodResult		*/ 	{ 	{ NULL, 0 }, 					{ NULL, 0 }, 						{ NULL, 0 },	 							{ NULL, 0 },											{ NULL, 0 }, 							{ ResetTrackerAndGoDefChannel, FSM_Idle } },	
/* EvaluationFailure		*/ 	{ 	{ NULL, 0 }, 					{ NULL, 0 }, 						{ NULL, 0 },	 							{ NULL, 0 },											{ NULL, 0 }, 							{ GoDefChannel, FSM_Idle } },
/* ForcedDownload			*/ 	{ 	{ NoAction, FSM_Searching }, 	{ NoAction, FSM_Searching }, 		{ NULL, 0 },	 							{ NULL, 0 },											{ NULL, 0 }, 							{ NULL, 0 } }
};

// 8. Initialize the Finite State Machine (FSM)
FSM_NEW(fsm_base, FSM_Base_States_Count, FSM_Base_Events_Count, fsm_base_transitions, fsm_base_timeouts);


void _ConnectToTracker() {
	LOGREMOTEANDLOCAL("Connecting to device id {%u} on channel {%u}...", _tracker_dl_status.TrackerNodeId, _tracker_dl_status.CommChannel);
	LogEngine_SaveText("Try: %u", _tracker_dl_status.TrackerNodeId);
	
	// request base station mode from the Tracker
	base_command_t bc;
	bc.BaseCommand = BaseMode;
	bc.Payload.BaseMode.Channel = _tracker_dl_status.CommChannel;
	
	network_payload_t p;
	Network_InitSendArgs(&p);
	p.Buffers[0].Size = sizeof(base_command_t);
	p.Buffers[0].Buffer = &bc;
	
	// try to retransmit 5 times
	Network_SendWithoutACK(_tracker_dl_status.TrackerNodeId, 101, &p);
	Timers_Block(4096);
	Network_SendWithoutACK(_tracker_dl_status.TrackerNodeId, 101, &p);
	Timers_Block(TICKS_SECOND);
	Network_SendWithoutACK(_tracker_dl_status.TrackerNodeId, 101, &p);
	Timers_Block(4096);
	Network_SendWithoutACK(_tracker_dl_status.TrackerNodeId, 101, &p);
	Timers_Block(TICKS_SECOND);
	Network_SendWithoutACK(_tracker_dl_status.TrackerNodeId, 101, &p);
	Timers_Block(4096);

	// follow to the selected channel
	Network_SetChannel(_tracker_dl_status.CommChannel);
	Network_PowerOn_RX();	
}

void Handler_Heartbeat(uint8_t source, const uint8_t* data) {
	// it is important to copy the data in memory so that data types alignment doesn't get corrupted.
	// not relevant for a series of uint8_t bytes, but very relevant for e.g. uint32_t
	logentry_t hb;
	memcpy(&hb, data, sizeof(logentry_t));
	
	// process only heartbeats
	if (hb.TypeId != Heartbeat)
		return;

	// store the hearbeat
	LogEngine_Save(&hb);

	// provide output for debugging
	char batteryAsString[BATTERY_READ_STRING_LENGTH];
	ADC_Devices_Battery_Convert(hb.Payload.Heartbeat.BatteryLevel, NULL, batteryAsString);
	
	LOGATT("Heartbeat from %u [ Uptime: %lu(%luh), LogSize: %lu, New Logs: %u, New Fix: %u, Voltage %s (%u) ]",
				source,
				hb.Payload.Heartbeat.Uptime,
				hb.Payload.Heartbeat.Uptime / 3600, // in hours
				hb.Payload.Heartbeat.LogSize,
				hb.Payload.Heartbeat.NewLogEntriesCount,
				hb.Payload.Heartbeat.NewFixCount,
				batteryAsString,
				hb.Payload.Heartbeat.BatteryLevel
			);
	
	/**
	 * now evaluate and deciede what to do with this heartbeat
	 */
	
	// only when in idle / waiting state.
	if (fsm_base.CurrentState != FSM_Idle)
		return;

	if (COMMUNICATION_PROTOCOL_VERSION != hb.Payload.Heartbeat.AppVersion) {
		LOGREMOTEANDLOCAL("Protocol version mismatch, ignoring node {%u}!", source);
		LogEngine_SaveText("API! %u", source);
		return;
	}
	
	if (LogEngine_IsFull()) {
		LOGERR("Log Storage is full. Ignoring Heartbeat.");
		return;
	}

	// only when the criteria are met, the heartbeat should be processed
	if (
			hb.Payload.Heartbeat.NewFixCount < THRESHOLD_NEW_FIX_COUNT
			&&
			hb.Payload.Heartbeat.NewLogEntriesCount < THRESHOLD_NEW_LOG_COUNT
		)
		return;
	
	// perform FSM transition and continue only on success
	if (false == FSM_ProcessEvent(&fsm_base, GoodHeartbeat))
		return;

	// initialize status
	memset(&_tracker_dl_status, 0, sizeof(_tracker_dl_status));
	// prepare to process the search for the tracker
	_pingtracker_remaining_calls = 10;

	// communication channel will be set to the value of the node id of this base station.
	uint8_t channel = Configuration.NodeID;

	// persist information about the selected device
	_tracker_dl_status.TrackerNodeId = source;
	_tracker_dl_status.CommChannel = channel;

	_tracker_dl_status.RemoteLogSize = hb.Payload.Heartbeat.LogSize;
	_tracker_dl_status.RemoteNewLogEntriesCount = hb.Payload.Heartbeat.NewLogEntriesCount;

	_ConnectToTracker();
}

void Handler_BaseStationCommand(uint8_t source, const uint8_t* data) {
	// it is important to copy the data in memory so that data types alignment doesn't get corrupted.
	// not relevant for a series of uint8_t bytes, but very relevant for e.g. uint32_t
	base_command_t bc;
	memcpy(&bc, data, sizeof(base_command_t));
	
	switch(bc.BaseCommand) {
		case PingResponse : {
			LOGREMOTEANDLOCAL("Connected to device id {%u} on channel {%u}", _tracker_dl_status.TrackerNodeId, _tracker_dl_status.CommChannel);
			LogEngine_SaveText("Conn: %u", _tracker_dl_status.TrackerNodeId);
			
			FSM_ProcessEvent(&fsm_base, TrackerFound);
			break;
		}
		
		default : {
			LOGDBG("Unexpected base station command {%u} from {%u}", bc.BaseCommand, source);
		}
	}
}

void FSM_Base_Init() {
	// register the radio handler so that heartbeats can be received by the application
	Network_RegisterHandler(100, Handler_Heartbeat);
	Network_RegisterHandler(101, Handler_BaseStationCommand);

	// prepare state machine init args
	fsm_startparameters_t init;
	
	init.FSM 			= &fsm_base;
	init.Delay 			= TICKS_60SECONDS;
	init.InitialState 	= FSM_Waiting;

	// start the fsm
	FSM_Start(&init);
}

bool FSM_Base_Abort() {
	if (fsm_base.CurrentState != FSM_Waiting) {
		LOGERR("Too late to abort!");
		return false;
	}
	
	FSM_Stop(&fsm_base);
	return true;
}


// implementation of actions:

void NotifyAutoBSRun() {
	LOGATT("Autonomous Base Station Initialized.");
}


void PingTracker() {
	// this action tries to connect to the selected tracker device.
	// it's triggered by a timeout, a counter needs to be used in order to abort
	// this possible endless loop.
	if (_pingtracker_remaining_calls == 0) {
		FSM_ProcessEvent(&fsm_base, TrackerNotFound);
		return;
	}
	_pingtracker_remaining_calls--;

	
	// the tracker is supposed to be waiting on another channel, the base station switched to this channel already
	
	// request a response from the tracker
	base_command_t bc;
	bc.BaseCommand = PingRequest;
	
	network_payload_t p;
	Network_InitSendArgs(&p);
	p.Buffers[0].Size = sizeof(base_command_t);
	p.Buffers[0].Buffer = &bc;
	
	Network_SendWithoutACK(_tracker_dl_status.TrackerNodeId, 101, &p);
	LOGDBG("ping...");
}


void _SendKeepAlive() {
	base_command_t bc;
	bc.BaseCommand = KeepAlive;
	
	network_payload_t p;
	Network_InitSendArgs(&p);
	p.Buffers[0].Size = sizeof(base_command_t);
	p.Buffers[0].Buffer = &bc;
	
	// try to retransmit 5 times
	Network_SendWithoutACK(_tracker_dl_status.TrackerNodeId, 101, &p);
	Timers_Block(2048);
	Network_SendWithoutACK(_tracker_dl_status.TrackerNodeId, 101, &p);
	Timers_Block(2048);
	Network_SendWithoutACK(_tracker_dl_status.TrackerNodeId, 101, &p);
	Timers_Block(2048);
	Network_SendWithoutACK(_tracker_dl_status.TrackerNodeId, 101, &p);
	Timers_Block(2048);
	Network_SendWithoutACK(_tracker_dl_status.TrackerNodeId, 101, &p);
	Timers_Block(4096);
}


/**
 * Holds the value of the logengine.sync counter. Used during divisions allows to detect the number of new entries loaded
 */
uint32_t _logengine_sync_entriessynced_copy;

void PerformDownloadNewData() {
	// make sure the tracker stays in the base station mode for another timeout (predefined to 5 mintues).
	_SendKeepAlive();
	
	// the tracker reported its total number of log entries and a number of new entries since last reset
	// it's safe to assume that the new data wasn't downloaded by that device yet.
	// perform a simple request for this data.
	LOGREMOTEANDLOCAL("Requesting newest data from {%u}", _tracker_dl_status.TrackerNodeId);

	// the timeout for the current state will be activated when this function exits. this is a good
	// moment to adapt the timeout to the time required for the request to complete
	// the formula:
	// 	input: known number of new log entries that will be requested
	//	input: approx. estimation of time required to download one entry in ticks = approx. 50
	//	input: safety extra timeout of 5 seconds
	//	output: number of log entries * time required for one + TICKS_5SECONDS
	fsm_base_timeouts[FSM_LoadingNew] =	(uint32_t) _tracker_dl_status.RemoteNewLogEntriesCount * 50 + TICKS_5SECONDS;
	LOGDBG("Timeout for FSM_LoadingNew set to %lu ticks (%lu seconds)", fsm_base_timeouts[FSM_LoadingNew], fsm_base_timeouts[FSM_LoadingNew] / TICKS_SECOND);
	
	// select the range properly. the last sequence number is basically the log size
	// the first is (logsize - new log entries)
	uint32_t firstSeqNo = 1;
	if (_tracker_dl_status.RemoteLogSize > _tracker_dl_status.RemoteNewLogEntriesCount)
		firstSeqNo = _tracker_dl_status.RemoteLogSize - _tracker_dl_status.RemoteNewLogEntriesCount + 1; // "+1" because we have the first seq no already 
		
	uint32_t lastSeqNo = _tracker_dl_status.RemoteLogSize;

	// create a copy so that number of loaded entries can be detected
	_logengine_sync_entriessynced_copy = LogEngine_Sync_EntriesSynced;
	
	uint32_t localSeqNo = LogEngine_Sync_Request(_tracker_dl_status.TrackerNodeId, &firstSeqNo, &lastSeqNo);
	
	_tracker_dl_status.LocalSequenceNo = localSeqNo;
	
	// the next step will be to download the missing data, prepare of this step
	_evaluatedownload_remaining_calls				= 4;
	_tracker_dl_status.ReceivedMissingCounter		= 0;
}

static gap_search_query_t query;

void _DownloadMissingDataCompleted() {
	// data about the recent run is available in the 'query' variable
	FSM_ProcessEvent(&fsm_base, DownloadPerformed);
}

void PerformDownloadMissingData() {
	
	// make sure the tracker stays in the base station mode for another timeout (predefined to 5 mintues).
	_SendKeepAlive();
	
	if (fsm_base.PreviousState == FSM_LoadingNew) {
		_tracker_dl_status.ReceivedNewDataCounter = LogEngine_Sync_EntriesSynced - _logengine_sync_entriessynced_copy;
		
		LOGREMOTEANDLOCAL("Received {%lu} of {%u} new entries from {%u}",
				_tracker_dl_status.ReceivedNewDataCounter,
				_tracker_dl_status.RemoteNewLogEntriesCount,
				_tracker_dl_status.TrackerNodeId);
		LogEngine_SaveText("DoneNewData");
	}
	
	LOGREMOTEANDLOCAL("Requesting missing data from {%u}", _tracker_dl_status.TrackerNodeId);
	
	memset(&query, 0, sizeof(gap_search_query_t));
	query.Request.LocalSeqNo				= 1; // search in the entire local memory
	query.Request.FirstSeqNo				= 1;
	query.Request.LastSeqNo					= _tracker_dl_status.RemoteLogSize;
	query.Request.TargetNodeId				= _tracker_dl_status.TrackerNodeId;
	query.Request.Handler_Sync_Completed	= _DownloadMissingDataCompleted;

	// create a copy so that number of loaded entries can be detected
	_logengine_sync_entriessynced_copy = LogEngine_Sync_EntriesSynced;
	
	LogEngine_Sync_FillGaps(&query);
}

void EvaluateDownload() {
	/**
	 * Information available:
	 *  - number of entries received for 'request'
	 *  - number of entries received for 'fillgaps'
	 *  - number of missing entries
	 *  - number of all entries expected to be there
	 * 
	 * How to evaluate:
	 *  if nothing was received at all -> repeat until counter reaches 0
	 * 		then, if the counter reaches 0 and nothing was received -> failure
	 *
	 * anytime:
	 *  if missing percentage less than 5, success 
	 */

	// keep track of data downloaded
	uint32_t progress = (LogEngine_Sync_EntriesSynced - _logengine_sync_entriessynced_copy);
	_tracker_dl_status.ReceivedMissingCounter += progress;
	
	// copy the number of items that were missing (keep the maximum, it's most interesting (it will be maximum only during the first run))
	if (query.Response.MissingCount > _tracker_dl_status.RequestedMissingCounter)
		_tracker_dl_status.RequestedMissingCounter = query.Response.MissingCount;
	
	LOGREMOTEANDLOCAL("Requested {%lu}, received {%lu} ({%lu} in total)", query.Response.MissingCount, progress, _tracker_dl_status.ReceivedMissingCounter);
	
	if (_evaluatedownload_remaining_calls == 0) {
		// few download calls were executed alredy, success only, when at least a bit was downloaded
		if (_tracker_dl_status.ReceivedNewDataCounter > 0 || _tracker_dl_status.ReceivedMissingCounter > 0) {
			FSM_ProcessEvent(&fsm_base, EvaluationGoodResult);
			return;
		}

		// this is the 'failure' path, no data was downloaded despite trying for few times
		FSM_ProcessEvent(&fsm_base, EvaluationFailure);
		return;
	}
	_evaluatedownload_remaining_calls--;
	
	uint32_t percentMissing = (100 * query.Response.MissingCount) / query.Response.ExpectedCount;
	
	if (percentMissing > THRESHOLD_MISSING_ENTRIES) {
		// room for improvement, continue the retry-loop
		FSM_ProcessEvent(&fsm_base, EvaluationRetryRequest);
		return;
	}
	
	// now, when the memory was scanned, less than 5% of data is missing
	// this is good, but in order to follow the success path, at least few 
	// log entries should have been transferred

	if (_tracker_dl_status.ReceivedNewDataCounter > 0 || _tracker_dl_status.ReceivedMissingCounter > 0) {
		LOGREMOTEANDLOCAL("Detected less than %u percent missing and got some data.", THRESHOLD_MISSING_ENTRIES);
		FSM_ProcessEvent(&fsm_base, EvaluationGoodResult);
	}
}


void GoDefChannel() {
	LOGREMOTEANDLOCAL("Returning to channel {0}");

	Network_SetChannel(0);
	Network_PowerOn_RX();


	if (fsm_base.PreviousState == FSM_Searching) {
		LOGREMOTEANDLOCAL("Lost device id {%u} on channel {%u}", _tracker_dl_status.TrackerNodeId, _tracker_dl_status.CommChannel);
		LogEngine_SaveText("Lost: %u", _tracker_dl_status.TrackerNodeId);
	} else {
		LogEngine_SaveText("Done: %u", _tracker_dl_status.TrackerNodeId);
	}
	
	LOGREMOTEANDLOCAL("Got {%lu}/{%u} new and {%lu}/{%lu} missing from {%u}.",
			_tracker_dl_status.ReceivedNewDataCounter,
			_tracker_dl_status.RemoteNewLogEntriesCount,
			_tracker_dl_status.ReceivedMissingCounter,
			_tracker_dl_status.RequestedMissingCounter,
			_tracker_dl_status.TrackerNodeId);

	memset(&_tracker_dl_status, 0, sizeof(_tracker_dl_status));
	
	LOGREMOTEANDLOCAL("Waiting for new trackers...");
}

void ResetTrackerAndGoDefChannel() {
	// request the tracker to reset
	base_command_t bc;
	bc.BaseCommand = ResetTracker;
	
	network_payload_t p;
	Network_InitSendArgs(&p);
	p.Buffers[0].Size = sizeof(base_command_t);
	p.Buffers[0].Buffer = &bc;
	
	// try to retransmit 5 times
	Network_SendWithoutACK(_tracker_dl_status.TrackerNodeId, 101, &p);
	Timers_Block(4096);
	Network_SendWithoutACK(_tracker_dl_status.TrackerNodeId, 101, &p);
	Timers_Block(4096);
	Network_SendWithoutACK(_tracker_dl_status.TrackerNodeId, 101, &p);
	Timers_Block(4096);
	Network_SendWithoutACK(_tracker_dl_status.TrackerNodeId, 101, &p);
	Timers_Block(4096);
	Network_SendWithoutACK(_tracker_dl_status.TrackerNodeId, 101, &p);
	
	GoDefChannel();
}

COMMAND(forcedl, "Forces a download from a selected tracker. Restarts the base station state machine.", cmd_args) {
	if (cmd_args->ArgumentCount != 3) {
		snprintf(cmd_args->ResponseString, cmd_args->ResponseStringMaxLength,
				"Help: forcedl [nodeid] [r.logsize] [r.newlogs]");
		return;
	}

	// perform FSM transition and continue only on success
	if (false == FSM_ProcessEvent(&fsm_base, ForcedDownload))
		return;
	
	// this will continue here also when the FSM was stoppped using the 'abort' command
	// stopping the FSM freezes its previous state and since aborting is only possible while FSM_Waiting
	// the FSM will just restart and continue working

	// initialize status
	memset(&_tracker_dl_status, 0, sizeof(_tracker_dl_status));
	// prepare to process the search for the tracker
	_pingtracker_remaining_calls = 10;

	// persist information about the selected device
	_tracker_dl_status.TrackerNodeId = atol(cmd_args->ArgumentArray[0]);
	_tracker_dl_status.CommChannel = Configuration.NodeID;
	
	_tracker_dl_status.RemoteLogSize = atoi(cmd_args->ArgumentArray[1]);
	_tracker_dl_status.RemoteNewLogEntriesCount = atol(cmd_args->ArgumentArray[2]);

	_ConnectToTracker();
}



#ifndef DOXYGEN_PUBLIC_DOC
/**
* Copyright (c) Microsoft Corporation. All rights reserved.
* Licensed under the Apache License, Version 2.0 (the "License"); you may not use these files except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
* Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.
 
* Copyright (c) 2012 Freie Universitaet Berlin, Tomasz Naumowicz. All rights reserved.
* 
* Redistribution and use in source and binary forms, with or without modification, are
* permitted provided that the following conditions are met:
* 
* - Redistributions of source code must retain the above copyright notice, this list of
* conditions and the following disclaimer.
* 
* - Redistributions in binary form must reproduce the above copyright notice, this list
* of conditions and the following disclaimer in the documentation and/or other materials
* provided with the distribution.
* 
* - Neither the name of the Freie Universitaet Berlin nor the
* names of its contributors may be used to endorse or promote products
* derived from this software without specific prior written permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Freie Universitaet Berlin BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#endif // DOXYGEN
