/**
 * @file		
 * @brief		@b Application: State machine controling the user radio behavior.
 * The \ref FSM.h contains the implementation of the Finite State Machine.
 * This file contains the description of the project specific FSM.
 */

#ifndef FSM_BASE_H_
#define FSM_BASE_H_

#include "Application.API.h"


/**
 * @brief	Starts the user radio state machine.
 * 
 * The fields highlighted in blue are definitions that can be modified in the \ref AppConfig.h file and influence
 * the behavior of the application. The green color of transitions highlights the most common path
 * during FSM processing.
 * 
 * The behavior of the state machine:
 * 
 * @dot
 * digraph fsm {
 * 		graph [fontname=Helvetica, fontsize=10];
 * 		edge [fontname=Helvetica, fontsize=10];
 * 
 *		node [shape=doublecircle, fontname=Helvetica, fontsize=10, fontstyle="bold"];
 *
 * 		wait						[label = "Waiting 60 seconds"]; 
 * 
 *		node [shape=circle, fontname=Helvetica, fontsize=10];
 * 
 * 		rx							[label = "RX (channel 0)", color="darkgreen", style="filled", fillcolor="lightgrey"];
 * 		search						[label = "Searching for Tracker\non Private Channel"];
 * 		request						[label = "Loading Newest Data"];
 * 		fillgaps					[label = "Loading Missing Data"];
 * 		eval						[label = "Evaluating Progress"];
 * 		stop						[label = "Stoppped\nRX on channel 0"]; 
 * 
 *		node [shape=note, fontname=Helvetica, fontsize=8];
 * 		evalSuccessNote				[label = "Less than or exactly THRESHOLD_MISSING_ENTRIES% missing, or\lRetried 5 times & received > 0\l", nojustify=true, fontcolor="blue"];
 * 		evalFailNote				[label = "Retried 5 times & received 0"];
 * 		rxToSearchNote				[label = "More than THRESHOLD_NEW_FIX_COUNT new fixes, or\lMore than THRESHOLD_NEW_LOG_COUNT new log entries\l", fontcolor="blue"];
 * 
 * 		wait -> rx					[label="Timeout\n(Start Autonomous Mode)", style="dashed", color="darkgreen"];
 * 		wait -> stop				[label="'abort'"];
 * 
 * 		stop -> search				[label="'forcedl'\n(Switch to Base Station mode\non private channel)"];
 * 		rx -> search				[label="'forcedl'\n(Switch to Base Station mode\non private channel)"];
 * 		wait -> search				[label="'forcedl'\n(Switch to Base Station mode\non private channel)"];
 * 		
 * 		rx -> rxToSearchNote		[label = "Heartbeat Received", dir = none, color="darkgreen"];
 * 		rxToSearchNote -> search	[label = "(Switch to Base Station mode\non private channel)", color="darkgreen"];
 * 
 * 		search -> search			[label = "Retry 10 times\n(ping)", style="dashed"];
 * 		search -> request			[label = "Tracker Found", color="darkgreen"];
 * 		search -> rx				[label = "Tracker Not Found\n(channel 0)"];
 * 
 * 		request -> fillgaps			[label = "Adaptive Timeout", style="dashed", color="darkgreen"];
 * 		fillgaps -> eval			[label = "Download Performed", color="darkgreen"];
 * 
 * 		eval -> fillgaps			[label = "Retry"];
 * 
 * 		eval -> evalSuccessNote		[dir = none, color="darkgreen"];
 * 		evalSuccessNote -> rx		[label = "Success\n(reset tracker, channel 0)", color="darkgreen"];
 * 
 * 		eval -> evalFailNote		[dir = none];
 * 		evalFailNote -> rx			[label = "Failure\n(channel 0)"];
 * }
 * @enddot
 */
void FSM_Base_Init();

/**
 * @brief	Aborts the FSM (available for a specified time after the power on event)
 */
bool FSM_Base_Abort();


#endif /*FSM_GPS_H_*/


#ifndef DOXYGEN_PUBLIC_DOC
/**
* Copyright (c) Microsoft Corporation. All rights reserved.
* Licensed under the Apache License, Version 2.0 (the "License"); you may not use these files except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
* Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.
 
* Copyright (c) 2012 Freie Universitaet Berlin, Tomasz Naumowicz. All rights reserved.
* 
* Redistribution and use in source and binary forms, with or without modification, are
* permitted provided that the following conditions are met:
* 
* - Redistributions of source code must retain the above copyright notice, this list of
* conditions and the following disclaimer.
* 
* - Redistributions in binary form must reproduce the above copyright notice, this list
* of conditions and the following disclaimer in the documentation and/or other materials
* provided with the distribution.
* 
* - Neither the name of the Freie Universitaet Berlin nor the
* names of its contributors may be used to endorse or promote products
* derived from this software without specific prior written permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Freie Universitaet Berlin BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#endif // DOXYGEN
