/**
 * @file		Application.h
 * @brief		User Application: Autonomous Tracker
 * Project specific user application for GPS tracking.
 * 
 * 
 * @section htdl How to download data:
 * 
 * Scenario:
 * - You are operating the base station (a node with the nodeid set to 1) using the serial protocol
 * - You installed this GPS tracker application to a device with a nodeid set to 56
 * - You used the \ref setgps \ref setradio and \ref suppress commands to configure the device 56
 * 
 * @subsection thdl1	1.	After receiving a heartbeat a transition to the base station should be requested
 * 
 * @code
 * @56 base 10
 * @endcode
 * 
 * 		where 56 is the target node and 10 is the communication channel (addresses in the range [0-127] are valid).
 * 
 *  	The node 56 confirms the request and you should be seeing the following message:
 * @code
 * [56:L] Switching to base station mode on channel {10}...
 * @endcode
 * 
 * @subsection thdl2	2. 	Switch to the same communication channel:
 * 	
 * @code
 * channel 10
 * @endcode
 * 
 * @subsection thdl3	3.	Optional: format your local log storage to make space for data download
 * 
 * @code
 * format
 * @endcode
 * 
 * @subsection thdl4	4.	Query the log size of the remote node
 * 
 * @code
 * @56 logsize
 * @endcode
 * 
 * 		You should receive a response:
 * 
 * @code
 * [56:logsize] {1358} slots used
 * @endcode
 * 
 * @subsection thdl5	5.	Load first batch of data in a sweep
 * 
 * @code
 * request 56 1 1358
 * @endcode
 * 
 * 		in this example following parameters were specified:
 * 			- 56	the target node
 * 			- 1		the sequence number of the first log entry that shall be downloaded
 * 			- 1358	the sequence number of the last log entry that shall be downloaded
 * 
 * 		You will see a confirmation of parameters accepted by the 'request' command:
 * 
 * @code
 * Local Sequence No:      1
 * Target NodeID:          56
 * First Sequence No:      1
 * Last Sequence No:       1358
 * [1:request] OK
 * @endcode
 * 
 * 		Now the donwload starts. In most releases of the firmware, a '.' will be printed for every log entry received
 * 		so that you can observe the progress of the data download.
 * 
 * 		Note that there is also one extra information available in the confirmation above, the "Local Sequence No".
 * 		It's relevant for further steps. It tells you that the data downloaded now starts at the "Local Sequence No"
 * 		in your local storage.
 * 
 *  @subsection thdl6	6.	Typically some data gets lost during transmission, this step can be repeated until all gaps in the data are filled.
 * 
 * 		Request a gap fill
 * 		
 * @code
 * fillgaps 1 56 1 1358
 * @endcode
 * 
 * 		This command starts a gap detect and fill process. In this example following parameters were used:
 * 			- 1		the local sequence number in the log from which the local search should start (all previous log entries are ignored)
 * 			- 56	the target node
 * 			- 1		the sequence number of the first log entry that shall be downloaded
 * 			- 1358	the sequence number of the last log entry that shall be downloaded
 * 
 * 		A series of operations will be scheduled:
 * 
 * @code
 * [1:L] Scheduled the operation. Wait for the schedule to complete!
 * (...)
 * [1:L] Processing remaining 1358 items...
 * @endcode
 * 
 * 		You need to wait until all operations complete.
 * 
 *  	Each operation produces similar status output:
 * 
 * @code
 * Local Sequence No:      1
 * Target NodeID:          56
 * First Sequence No:      417
 * Last Sequence No:       832
 * Missing sequence numbers (searching...): 423 431 445 453 474 482 503 511 532 540 561 569 583 591 612 620 641 649 670 678 699 707 721 729 750 758 779 787 808 816
 * Detected missing:       7% (30/416)
 * @endcode
 * 
 * 		In brief this status report describes the current task, the range it's processing and the sequence numbers of missing entries
 * 		in the investigated range. A percentage of missing entries is also given.
 * 
 * @remarks 
 * - You can't abort a scheduled fillgaps series. You need to wait until it terminates or reset the device.
 * - It's a good rule to stop scheduling further fillgaps operations when the percentage of missing entries is less than 10%
 * 
 * @subsection thdl7	7. Removing data from the remote device
 * 
 * 		After succesfuly downloading data from the remote device, you can request a reset of the device and also remove the data
 * 		stored remotely
 * 
 * 		This command requests a format of the storage and a reset of the device (reset restarts the tracking application)
 * @code
 * @56 f_reset
 * @endcode
 *
 * 		The remote node confirms the request:
 * @code 
 * [56:L] Formatting and restarting the device... (1)
 * @endcode
 * 
 * @remarks
 * 		Remember to reconfigure your base station back to channel 0 using the "channel 0" command.
 */

#ifndef APPLICATION_H_
#define APPLICATION_H_

#include "AppConfig.h"

void Application_Init(); 

#endif /*APPLICATION_H_*/

#ifndef DOXYGEN_PUBLIC_DOC
/**
* Copyright (c) Microsoft Corporation. All rights reserved.
* Licensed under the Apache License, Version 2.0 (the "License"); you may not use these files except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
* Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.
 
* Copyright (c) 2012 Freie Universitaet Berlin, Tomasz Naumowicz. All rights reserved.
* 
* Redistribution and use in source and binary forms, with or without modification, are
* permitted provided that the following conditions are met:
* 
* - Redistributions of source code must retain the above copyright notice, this list of
* conditions and the following disclaimer.
* 
* - Redistributions in binary form must reproduce the above copyright notice, this list
* of conditions and the following disclaimer in the documentation and/or other materials
* provided with the distribution.
* 
* - Neither the name of the Freie Universitaet Berlin nor the
* names of its contributors may be used to endorse or promote products
* derived from this software without specific prior written permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Freie Universitaet Berlin BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#endif // DOXYGEN
