/**
 * @file		FSM_Acc.c
 * @ingroup		FSM definitions
 * @brief		Accelerometer sensor sampling FSM
 * The CMA is a three-axis accelerometer developed by VTI (http://www.vti.fi/en/support/).
 * 
 * 
 */

#include "Application.API.h"

#include "FSM.h"
#include "FSM_Acc.h"

// 1. Define all actions here.
//    This could happen in the header file, but you might like to keep those functions local.
//    This definition is required for the definition of the transition table

void NoActionAcc(){}				// no action
void ACC_Measurement();				// enable and measure Acc data for a period
uint16_t NoOfRemainingLogs = 0;
void SetNoOfRemainingLogs();

// 2. Define states you are going to use in your state machine
enum FSM_Acc_States {
	FSM_Idle 					= 0,
	FSM_Suspended				= 1,
	FSM_Measurement 			= 2,
	FSM_MeasInterval			= 3
};

// 3. Count your states
#define FSM_Acc_States_Count		4

// 4. Define events used in your state machine
//    Leave 'Timeout' at the beginning
enum FSM_Acc_Events {
	ACCFSM_Timeout				= 0,
	ACCFSM_LoggedMeasurement	= 1,
	ACCFSM_Suspend				= 2
};

// 5. Count your events
#define FSM_Acc_Events_Count 	3


// 6. Define timeouts for your states
//    When no event happens within the timeout value, the state change will be executed (signalled as timeout)
//    0 stands for no timeout.
//    The order must correspond to the order in your states definition
uint32_t fsm_acc_timeouts[FSM_Acc_States_Count] = {
	TICKS_3MINUTES,
	TICKS_10MINUTES,
	TICKS_5SECONDS,
	204
};

// 7. Define your state transitions.
//    Horiziontal:	current state
//    Vertical:		The event
//    Entry:		Ation to call and the next state.
 FSM_Entry fsm_acc_transitions[FSM_Acc_Events_Count][FSM_Acc_States_Count] = {

//						Idle									Suspended						Measurement							MeasInterval		
/* Timeout  */ 		{	{ ACC_Measurement, FSM_Measurement }, 	{ NoActionAcc, FSM_Idle },		{ NoActionAcc, FSM_Idle },			{ ACC_Measurement, FSM_Measurement }},
/* LoggedM... */	{	{ NULL, 0 }, 							{ NULL, 0 },					{ NoActionAcc, FSM_MeasInterval },	{ NULL, 0 }},
/* Suspend */ 		{	{ CMA_Disable, FSM_Suspended }, 		{ NoActionAcc, FSM_Suspended },	{ CMA_Disable, FSM_Suspended },		{ CMA_Disable, FSM_Suspended }  }

};
 
// 8. Initialize the Finite State Machine (FSM)
FSM_NEW(fsm_acc, FSM_Acc_States_Count, FSM_Acc_Events_Count, fsm_acc_transitions, fsm_acc_timeouts);


// 9. Implement the init function and your actions
void ACC_Measurement() {

	//set the value of NoOfRemainingLogs if it is reset from the previous measurement phase
	if(NoOfRemainingLogs == 0)
		SetNoOfRemainingLogs();

	//CMA_Enable(true);
	uint8_t x, y, z;
	rawtime_t rt;
	//LOGDBG("ACC msrmnt: NoOfRemainingLogs: %u", NoOfRemainingLogs);
	
	logentry_t acc;
    LogEngine_PrepareLogEntry(&acc);
    acc.TypeId = ACC;
    
	if (CMA_Read(&x, &y, &z, &rt)) {
    	acc.Payload.ACC.AccX = x;
    	acc.Payload.ACC.AccY = y;
		acc.Payload.ACC.AccZ = z;
    	acc.Timestamp = rt;
    	//LOGDBG("ACC msrmnt: %u\t%u\t%u\t%u", x, y, z, rt);
    	LogEngine_Save(&acc);
		NoOfRemainingLogs--;
	}
	//trig event to the next state; either sleeping (Idle) or measurement period (MeasInterval)
	if(NoOfRemainingLogs > 0)
	{
		//LOGDBG("ACC msrmnt: LoggedMeasurement");
		FSM_ProcessEvent(&fsm_acc, ACCFSM_LoggedMeasurement);
	}
	else
	{
		//LOGDBG("ACC msrmnt: Timeout");
		FSM_ProcessEvent(&fsm_acc, ACCFSM_Timeout);
	}
}

static void _UpdateConfiguration() {
	if (false == AppConfiguration.Acc_Configured) {
		AppConfiguration.Acc_PowerUpDelay			= DEFAULT_ACC_POWERUP_DELAY;
		AppConfiguration.Acc_IdleTime 				= DEFAULT_ACC_IDLE_TIME;
		AppConfiguration.Acc_LoggingTime		 	= DEFAULT_ACC_LOGGING_TIME;
		AppConfiguration.Acc_SamplingInterval	 	= DEFAULT_ACC_SAMPLING_INTERVAL;
		AppConfiguration.Acc_SamplingRate			= DEFAULT_ACC_SAMPLINGRATE;
		AppConfiguration.Acc_gRange					= DEFAULT_ACC_GRANGE;
		
		AppConfiguration.Acc_Configured 	= true; // since default values were loaded
	}
	
	if (false == AppConfiguration.Suppress_Acc_Configured) {
		AppConfiguration.Suppress_Acc_Activation	= DEFAULT_SUPPRESS_ACC_ACTIVATION;
		AppConfiguration.Suppress_Acc_Deactivation 	= DEFAULT_SUPPRESS_ACC_DEACTIVATION;
		
		AppConfiguration.Suppress_Acc_Configured 	= true; // since the default values were loaded
	}
	
	fsm_acc_timeouts[FSM_Idle] 				= AppConfiguration.Acc_IdleTime;
	fsm_acc_timeouts[FSM_Measurement] 		= AppConfiguration.Acc_LoggingTime;
	fsm_acc_timeouts[FSM_MeasInterval] 		= AppConfiguration.Acc_SamplingInterval;

	SetNoOfRemainingLogs();
}

void SetNoOfRemainingLogs(){
	/*LOGDBG("Set: fsm_acc_timeouts %lu\t%lu\t%lu", 
		 fsm_acc_timeouts[FSM_Idle], fsm_acc_timeouts[FSM_Measurement], fsm_acc_timeouts[FSM_MeasInterval]);*/
	/*LOGDBG("Set: TICKS_SECOND, SamplingRate, SamplingInterval %lu\t%u\t%lu", 
		 TICKS_SECOND, AppConfiguration.Acc_SamplingRate, AppConfiguration.Acc_SamplingInterval);*/
	//save how many logs should be logged before measurement goes back to sleep
	NoOfRemainingLogs = (uint16_t)(AppConfiguration.Acc_LoggingTime / AppConfiguration.Acc_SamplingInterval);
	/*LOGDBG("Set: LoggingTime, SamplingInterval, RemainingLogs: %lu\t%lu\t%u", 
		AppConfiguration.Acc_LoggingTime, AppConfiguration.Acc_SamplingInterval, NoOfRemainingLogs);*/
}

void FSM_Acc_Init() {
	// process the configuration
	_UpdateConfiguration();
	
	//LOGDBG("Registering ACC Handler..");
	//CMA_Enable(true);
	//Handler_ACC_Results = Handler_ACC;
	
	fsm_startparameters_t init;
	init.FSM			= &fsm_acc;
	init.Delay			= AppConfiguration.Acc_PowerUpDelay;
	init.InitialState	= FSM_Idle;
	
    LOGDBG("Ends with FSM Start..");
	FSM_Start(&init);
}

void FSM_Acc_Abort() {
	FSM_Stop(&fsm_acc);
	CMA_Disable();
}

void FSM_Acc_Suspend() {
	FSM_ProcessEvent(&fsm_acc, ACCFSM_Suspend);
}

COMMAND(setacc, "Configures the ACC experiment", args) {
	if (args->ArgumentCount != 3) {
		printf("Help: setacc [power up delay] [sleep time between measurements] [measurement logging time]\r\n"
			   "Remark: 0 stands for infinite\r\n");

		printf("Power up delay:\t%lus\r\n"
			   "Sleep time:\t%lus\r\n"
			   "Logging:\t%lus\r\n",
			   
			   AppConfiguration.Acc_PowerUpDelay 	/ TICKS_SECOND,
			   AppConfiguration.Acc_IdleTime 		/ TICKS_SECOND,
			   AppConfiguration.Acc_LoggingTime 	/ TICKS_SECOND
			   
			   );
		
		snprintf(args->ResponseString, args->ResponseStringMaxLength,
				
				"Current configuration: {%lu,%lu,%lu}",
				
				AppConfiguration.Acc_PowerUpDelay 	/ TICKS_SECOND,
				AppConfiguration.Acc_IdleTime 		/ TICKS_SECOND,
				AppConfiguration.Acc_LoggingTime 	/ TICKS_SECOND
				
				);
		
		
		return;
	}
	
	// parameters count is ok, update the configuration
	AppConfiguration.Acc_PowerUpDelay		= (uint32_t) atol(args->ArgumentArray[0]) * TICKS_SECOND;
	AppConfiguration.Acc_IdleTime			= (uint32_t) atol(args->ArgumentArray[1]) * TICKS_SECOND;
	AppConfiguration.Acc_LoggingTime		= (uint32_t) atol(args->ArgumentArray[2]) * TICKS_SECOND;
	
	AppConfiguration.Acc_Configured		= true;
	
	Configuration_Save();
	
	SetNoOfRemainingLogs();

	snprintf(args->ResponseString, args->ResponseStringMaxLength,
			
			"{%lu,%lu,%lu} Changes saved, reset to activate.",
			
			AppConfiguration.Acc_PowerUpDelay 	/ TICKS_SECOND,
			AppConfiguration.Acc_IdleTime 		/ TICKS_SECOND,
			AppConfiguration.Acc_LoggingTime 	/ TICKS_SECOND
			
			);
}

COMMAND(acc, "ACC debug interface", cmd_args) {
	if (cmd_args->ArgumentCount != 1) {
		printf("Accelerometer measurement mode: Sampling Rate: %u Hz, Measurement Range: %u g\n",
			AppConfiguration.Acc_SamplingRate, AppConfiguration.Acc_gRange);
		printf("Usage: acc [Sampling Rate (Hz) and Measurement Range (g)]\r\n"
			" 1:   {40,8}\r\n"
			" 2:   {100,8}\r\n"
			" 3:   {400,8}\r\n"
			" 4:   {100,2}\r\n"
			" 5:   {400,2}\r\n"
			);
		return;
	}
	
	 switch (cmd_args->ArgumentArray[0][0]) {
		 case '1' : {
			 	AppConfiguration.Acc_SamplingRate	= 40;
				AppConfiguration.Acc_gRange			= 8;
				break;
		 }
		 case '2' : {
			 	AppConfiguration.Acc_SamplingRate	= 100;
				AppConfiguration.Acc_gRange			= 8;
				break;
		 }
		 case '3' : {
			 	AppConfiguration.Acc_SamplingRate	= 400;
				AppConfiguration.Acc_gRange			= 8;
			 	break;
		 }
		 case '4' : {
			 	AppConfiguration.Acc_SamplingRate	= 100;
				AppConfiguration.Acc_gRange			= 2;
				break;
		 }
		 
		 case '5' : {
			 	AppConfiguration.Acc_SamplingRate	= 400;
				AppConfiguration.Acc_gRange			= 2;
				break;
		 }
		 
		 default : {
			 // error
		 }
	 }
	 
	 //update sampling interval timeout with updated sampling rate
	 AppConfiguration.Acc_SamplingInterval = (uint32_t)(TICKS_SECOND / AppConfiguration.Acc_SamplingRate);	//this is in sec. and in TICKS_SECOND
	 /*LOGDBG("acc: TICKS_SECOND, SamplingRate, SamplingInterval %lu\t%u\t%lu", 
		 TICKS_SECOND, AppConfiguration.Acc_SamplingRate, AppConfiguration.Acc_SamplingInterval);*/

	 if(CMA_SetMode(AppConfiguration.Acc_SamplingRate, AppConfiguration.Acc_gRange)){
		AppConfiguration.Acc_Configured	= true;
	
		Configuration_Save();
	
		snprintf(cmd_args->ResponseString, cmd_args->ResponseStringMaxLength,
				"{%u,%u} Changes saved, reset to activate.",
				AppConfiguration.Acc_SamplingRate,
				AppConfiguration.Acc_gRange		
				);
		}
}

/*COMMAND(suppAcc, "Configures down time for the ACC (overrides setacc)", args) {
	if (args->ArgumentCount != 2) {
		printf("Help: suppress [HH of suppress activation] [HH of suppress deactivation]\r\n"
			   "Remark:\t suppress 10 13 means the ACC will be disabled when time >= 10 and time < 13\r\n"
			   "\t resulting in ACC activity from 10:00:00 till 12:59:59\r\n"
				"\t suppress has no effect when the time isn't correct\r\n"
				"\t suppress 10 10 disables this feature (when HH are equal the condition above is always false\r\n");

		if (AppConfiguration.Suppress_Acc_Configured) {
			snprintf(args->ResponseString, args->ResponseStringMaxLength,
					
					"Current configuration: {%u,%u}",
					
					AppConfiguration.Suppress_Acc_Activation,
					AppConfiguration.Suppress_Acc_Deactivation);
		}
		
		return;
	}
	
	AppConfiguration.Suppress_Acc_Activation	= atoi(args->ArgumentArray[0]);
	AppConfiguration.Suppress_Acc_Deactivation 	= atoi(args->ArgumentArray[1]);
	AppConfiguration.Suppress_Acc_Configured 	= true;
	
	Configuration_Save();
	
	snprintf(args->ResponseString, args->ResponseStringMaxLength,
			
			"{%u,%u} Changes saved, reset to activate.",
			
			AppConfiguration.Suppress_Acc_Activation,
			AppConfiguration.Suppress_Acc_Deactivation
			
			);
}*/

#ifndef DOXYGEN_PUBLIC_DOC
/**
* Copyright (c) Microsoft Corporation. All rights reserved.
* Licensed under the Apache License, Version 2.0 (the "License"); you may not use these files except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
* Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.
*/
#endif // DOXYGEN
