#include "Application.API.h"

#include "FSM.h"
#include "FSM_Tracking.h"
#include "FSM_Radio.h"

#include "../Base/Shared/SharedTypes.h"

// 1. Define all actions here.
//    This could happen in the header file, but you might like to keep those functions local.
//    This definition is required for the definition of the transition table

void Heartbeat_EnableRX();
void DisableRX();

void SuspendGPS();

// 2. Define states you are going to use in your state machine
enum FSM_Radio_States {
	FSM_Idle 				= 0,
	FSM_Waiting				= 1,
	FSM_Chatting 			= 2
};

// 3. Count your states
#define FSM_Radio_States_Count	3

// 4. Define events used in your state machine
//    Leave 'Timeout' at the beginning
enum FSM_Radio_Events {
	Timeout					= 0,
	GotBaseStationRequest	= 1
};

// 5. Count your events
#define FSM_Radio_Events_Count 	2

// 6. Define timeouts for your states
//    When no event happens within the timeout value, the state change will be executed (signalled as timeout)
//    0 stands for no timeout.
uint32_t fsm_radio_timeouts[FSM_Radio_States_Count] = {
		TICKS_5MINUTES,
		TICKS_10SECONDS,
		TICKS_10MINUTES
};

// 7. Define your state transitions.
//    Horiziontal:	current state
//    Vertical:		The event
//    Entry:		Ation to call and the next state.
 FSM_Entry fsm_radio_transitions[FSM_Radio_Events_Count][FSM_Radio_States_Count] = {

/*						Idle									waiting							chatting		*/
/* Timeout  */ 		{ 	{ Heartbeat_EnableRX, FSM_Waiting }, 	{ DisableRX, FSM_Idle, }, 		{ DisableRX, FSM_Idle } },
/* got request */ 	{ 	{ SuspendGPS, FSM_Chatting }, 			{ SuspendGPS, FSM_Chatting }, 	{ SuspendGPS, FSM_Chatting } }
};

// 8. Initialize the Finite State Machine (FSM)
FSM_NEW(fsm_radio, FSM_Radio_States_Count, FSM_Radio_Events_Count, fsm_radio_transitions, fsm_radio_timeouts);

void Handler_Heartbeat(uint8_t source, const uint8_t* data) {
	// it is important to copy the data in memory so that data types alignment doesn't get corrupted.
	// not relevant for a series of uint8_t bytes, but very relevant for e.g. uint32_t
	logentry_t hb;
	memcpy(&hb, data, sizeof(logentry_t));
	
	if (hb.TypeId != Heartbeat)
		return;

	// store the hearbeat
	LogEngine_Save(&hb);

	/*
	char batteryAsString[BATTERY_READ_STRING_LENGTH];
	ADC_Devices_Battery_Convert(hb.Payload.Battery, NULL, batteryAsString);
	
	LOGATT("Heartbeat from %u [ Uptime: %lu(%luh), LogSize: %lu, FixCount: %lu, Voltage %s (%u) ]",
				source,
				hb.Uptime,
				hb.Uptime / 3600, // in hours
				hb.LogSize,
				hb.FixCount,
				batteryAsString,
				hb.BatteryVoltage
			);
	*/
}

bool ProcessBaseModeRequest(uint16_t requestedChannel) {
	bool processed = FSM_ProcessEvent(&fsm_radio, GotBaseStationRequest);
	
	if (processed) {
		
		LogEngine_SaveText("BS mode");
		
		LOGREMOTE("Switching to base station mode on channel {%u}...", requestedChannel);
		LOGREMOTE("Switching to base station mode on channel {%u}...", requestedChannel);
		LOGREMOTE("Switching to base station mode on channel {%u}...", requestedChannel);
		
		Network_SetChannel(requestedChannel);
		
		Network_PowerOn_RX();
	}
	
	return processed;
}

void Handler_BaseStationCommand(uint8_t source, const uint8_t* data) {
	// it is important to copy the data in memory so that data types alignment doesn't get corrupted.
	// not relevant for a series of uint8_t bytes, but very relevant for e.g. uint32_t
	base_command_t bc;
	memcpy(&bc, data, sizeof(base_command_t));
	
	switch(bc.BaseCommand) {
		case BaseMode : {
			ProcessBaseModeRequest(bc.Payload.BaseMode.Channel);
			break;
		}
		case PingRequest : {
			// reply to the base
			base_command_t bc;
			bc.BaseCommand = PingResponse;
			
			network_payload_t p;
			Network_InitSendArgs(&p);
			p.Buffers[0].Size = sizeof(base_command_t);
			p.Buffers[0].Buffer = &bc;
			
			Network_SendWithoutACK(source, 101, &p);
			
			break;
		}
		case ResetTracker : {
			Watchdog_DeviceReset();			
			break;
		}
		case KeepAlive : {
			// extend the timeout so that the base station mode doesn't stop too early
			// it's enought to ask the fsm to process the base station command
			FSM_ProcessEvent(&fsm_radio, GotBaseStationRequest);
			
			break;
		}
		
		default : {
			LOGDBG("Unexpected base station command {%u} from {%u}", bc.BaseCommand, source);
		}
	}
}




int16_t _blinkledtimerid;

void _BlinkLED_ledoff(uint16_t notUsed) {
	LED_OFF;
}

void _BlinkLED(uint16_t notUsed) {
	LED_ON;
	Timers_Add(_BlinkLED_ledoff, 256, 0);
	
	_blinkledtimerid = Timers_Add(_BlinkLED, TICKS_SECOND, 0);
}

void _BlinkLED_Abort() {
	Timers_Remove(_blinkledtimerid);
	LED_OFF;
}

static void _UpdateConfiguration() {
	if (false == AppConfiguration.Radio_Configured) {
		AppConfiguration.Radio_IdleTime 		= DEFAULT_RADIO_IDLE_TIME;
		AppConfiguration.Radio_WaitForBaseTime 	= DEFAULT_RADIO_WAIT_FOR_BASE_TIME;
		AppConfiguration.Radio_PowerUpDelay		= DEFAULT_RADIO_POWERUP_DELAY;
		
		AppConfiguration.Radio_Configured 		= true;
	}
	
	fsm_radio_timeouts[FSM_Idle]	= AppConfiguration.Radio_IdleTime;
	fsm_radio_timeouts[FSM_Waiting]	= AppConfiguration.Radio_WaitForBaseTime;
}

void FSM_Radio_Init() {
	// process the configuration
	_UpdateConfiguration();
	
	fsm_startparameters_t init;
	init.FSM = &fsm_radio;
	init.Delay = AppConfiguration.Radio_PowerUpDelay;
	init.InitialState = FSM_Idle;
	
	Network_RegisterHandler(100, Handler_Heartbeat);
	Network_RegisterHandler(101, Handler_BaseStationCommand);
	
	FSM_Radio_InitLogSize = LogEngine_GetLogSize();
	
	FSM_Start(&init);
}

void FSM_Radio_Abort() {
	FSM_Stop(&fsm_radio);
}


void Heartbeat_EnableRX() {
	// generate hartbeat information	
	logentry_t hb;
	LogEngine_PrepareLogEntry(&hb);
	hb.TypeId = Heartbeat;
	
	// set the RFC number to 1149
	hb.Payload.Heartbeat.RFC					= 1149;

	hb.Payload.Heartbeat.LogSize 				= LogEngine_GetLogSize();
	hb.Payload.Heartbeat.Uptime 				= (*TICKS_SINCE_POR() >> 13);
	hb.Payload.Heartbeat.BatteryLevel 			= FSM_Tracking_LastVoltage;
	hb.Payload.Heartbeat.NewFixCount 			= FSM_Tracking_FixCounter;	
	hb.Payload.Heartbeat.NewLogEntriesCount		= (uint16_t) (hb.Payload.Heartbeat.LogSize - FSM_Radio_InitLogSize);

	// set the protocol version
	hb.Payload.Heartbeat.AppVersion				= COMMUNICATION_PROTOCOL_VERSION;
	
	// transmit the heartbeat
	network_payload_t p;
	Network_InitSendArgs(&p);
	p.Buffers[0].Buffer = &hb;
	p.Buffers[0].Size = sizeof (logentry_t);
	
	Network_SendWithoutACK(0, 100, &p);
	Timers_Block(8192);
	Network_SendWithoutACK(0, 100, &p);
	Timers_Block(8192);
	Network_SendWithoutACK(0, 100, &p);
	// transmitting exits in RX mode
}

void DisableRX() {
	// the chat timed out. but the device still might be in another channel
	// go back to the default channel
	Network_SetChannel(0);
	
	// now disable the radio
	Network_PowerOff();
	
	if (fsm_radio.CurrentState != FSM_Chatting) {
		// this is the timeout transition, the LED2 should be disabled
		_BlinkLED_Abort();
	}
}

void SuspendGPS() {
	LOGDBG("Entered Base Station mode.");
	
	if (fsm_radio.PreviousState != FSM_Chatting) {
		_BlinkLED(0);
	}
	
	FSM_Tracking_Suspend();
}

COMMAND(setradio, "Configures the radio experiment", args) {
	if (args->ArgumentCount != 3) {
		printf("Help: setradio [power up delay] [sleep time between RX] [max time to wait for base]\r\n"
			   "Remark: 0 stands for infinite, executing 'setradio xx 0 1' disables the radio xx seconds after power on.\r\n");

		printf("Power up delay:\t%lus\r\n"
			   "Sleep time:\t%lus\r\n"
			   "Max. RX:\t%lus\r\n",
			   
			   AppConfiguration.Radio_PowerUpDelay 		/ TICKS_SECOND,
			   AppConfiguration.Radio_IdleTime			/ TICKS_SECOND,
			   AppConfiguration.Radio_WaitForBaseTime	/ TICKS_SECOND
			   
			   );
		
		
		snprintf(args->ResponseString, args->ResponseStringMaxLength,
				
					"Current configuration: {%lu,%lu,%lu}",
					
					AppConfiguration.Radio_PowerUpDelay 		/ TICKS_SECOND,
					   AppConfiguration.Radio_IdleTime 			/ TICKS_SECOND,
					   AppConfiguration.Radio_WaitForBaseTime 	/ TICKS_SECOND
					   
				);		
		return;
	}
	
	// parameters count is ok, update the configuration;
	AppConfiguration.Radio_PowerUpDelay			= (uint32_t) atol(args->ArgumentArray[0]) * TICKS_SECOND;
	AppConfiguration.Radio_IdleTime				= (uint32_t) atol(args->ArgumentArray[1]) * TICKS_SECOND;
	AppConfiguration.Radio_WaitForBaseTime		= (uint32_t) atol(args->ArgumentArray[2]) * TICKS_SECOND;
	
	AppConfiguration.Radio_Configured			= true;
	
	Configuration_Save();
	
	snprintf(args->ResponseString, args->ResponseStringMaxLength,
			
			"{%lu,%lu,%lu} Changes saved, reset to activate.",
			
			AppConfiguration.Radio_PowerUpDelay 	/ TICKS_SECOND,
			AppConfiguration.Radio_IdleTime 		/ TICKS_SECOND,
			AppConfiguration.Radio_WaitForBaseTime 	/ TICKS_SECOND
			
			);
}


COMMAND(base, "Brings the device to the 'base station' mode in the selected channel", args) {
	uint16_t channel = 0;
	if (args->ArgumentCount == 1) {
		channel = atoi(args->ArgumentArray[0]);
	}
	
	bool processed = ProcessBaseModeRequest(channel);
	
	if (processed) {
		snprintf(args->ResponseString, args->ResponseStringMaxLength, "at channel {%u}", channel);
	}
	else {
		snprintf(args->ResponseString, args->ResponseStringMaxLength, "ERROR!");
	}
}


#ifndef DOXYGEN_PUBLIC_DOC
/**
* Copyright (c) Microsoft Corporation. All rights reserved.
* Licensed under the Apache License, Version 2.0 (the "License"); you may not use these files except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
* Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.
 
* Copyright (c) 2012 Freie Universitaet Berlin, Tomasz Naumowicz. All rights reserved.
* 
* Redistribution and use in source and binary forms, with or without modification, are
* permitted provided that the following conditions are met:
* 
* - Redistributions of source code must retain the above copyright notice, this list of
* conditions and the following disclaimer.
* 
* - Redistributions in binary form must reproduce the above copyright notice, this list
* of conditions and the following disclaimer in the documentation and/or other materials
* provided with the distribution.
* 
* - Neither the name of the Freie Universitaet Berlin nor the
* names of its contributors may be used to endorse or promote products
* derived from this software without specific prior written permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Freie Universitaet Berlin BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#endif // DOXYGEN
