/**
 * @file		FSM_Radio.h
 * @brief		State machine controling the user radio behavior.
 * The \ref FSM.h contains the implementation of the Finite State Machine.
 * This file contains the description of the project specific FSM.
 */

#ifndef FSM_RADIO_H_
#define FSM_RADIO_H_

#include "Application.API.h"


/**
 * Starts the user radio state machine.
 * 
 * The behavior of the state machine:
 * 
 * @dot
 * digraph fsm {
 * 		graph					[fontname=Helvetica, fontsize=10];
 * 		edge					[fontname=Helvetica, fontsize=10];
 * 
 *		node					[shape=doublecircle, fontname=Helvetica, fontsize=10];
 * 		wait					[label = "Waiting\nRX on Channel 0"];
 *  
 *		node					[shape=circle, fontname=Helvetica, fontsize=10];
 * 		idle					[label = "Idle\nRadio Disabled", color="darkgreen", style="filled", fillcolor="lightgrey"];
 *		waiting					[label = "Waiting for Base Station\non channel 0"];
 * 		chatting				[label = "Base Station Mode\nPrivate Channel"];
 * 
 * 		stop					[label = "Stoppped\nRX on channel 0"]; 
 * 
 * 		wait -> idle 			[label="Configurable Timeout\n(Start Tracking Mode)", style="dashed", color="darkgreen", fontcolor="blue"];
 * 		wait -> stop			[label="'abort'"];
 * 		idle -> waiting 		[label="Configurable Delay\n(Send Heartbeats)", style="dashed", color="darkgreen", fontcolor="blue"];
 * 		wait -> chatting 		[label="Base Station Request\n(Suspend GPS)"];
 * 		waiting -> idle 		[label="Configurable Timeout\n(Disable Radio)", style="dashed", color="darkgreen", fontcolor="blue"];
 * 		waiting -> chatting 	[label="Base Station Request\n(Suspend GPS)"];
 * 		chatting -> idle 		[label="10 Minutes Timeout\n(Disable Radio)", style="dashed"];
 * 		chatting -> chatting 	[label="Base Station Request, or\n\"Keep Alive\"\n(Suspend GPS)"];
 * 		stop -> chatting 		[label="Base Station Request\n(Suspend GPS)"];
 * }
 * @enddot
 */
void FSM_Radio_Init();

void FSM_Radio_Abort();

/**
 * Number of new logentries when previous heartbeat was sent.
 */
uint32_t	FSM_Radio_InitLogSize;

#endif /*FSM_GPS_H_*/

#ifndef DOXYGEN_PUBLIC_DOC
/**
* Copyright (c) Microsoft Corporation. All rights reserved.
* Licensed under the Apache License, Version 2.0 (the "License"); you may not use these files except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
* Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.
 
* Copyright (c) 2012 Freie Universitaet Berlin, Tomasz Naumowicz. All rights reserved.
* 
* Redistribution and use in source and binary forms, with or without modification, are
* permitted provided that the following conditions are met:
* 
* - Redistributions of source code must retain the above copyright notice, this list of
* conditions and the following disclaimer.
* 
* - Redistributions in binary form must reproduce the above copyright notice, this list
* of conditions and the following disclaimer in the documentation and/or other materials
* provided with the distribution.
* 
* - Neither the name of the Freie Universitaet Berlin nor the
* names of its contributors may be used to endorse or promote products
* derived from this software without specific prior written permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Freie Universitaet Berlin BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#endif // DOXYGEN
