#include "Application.API.h"

#include "FSM.h"
#include "FSM_Tracking.h"

// 1. Define all actions here.
//    This could happen in the header file, but you might like to keep those functions local.
//    This definition is required for the definition of the transition table

void NoAction(){}					// no action
void EnableGPS(); 					// enables gps sensing
void DisableGPS(); 					// disables gps sensing

// 2. Define states you are going to use in your state machine
enum FSM_States {
	FSM_Idle 					= 0,
	FSM_WaitingForFix 			= 1,
	FSM_WaitingForRMC			= 2,
	FSM_Suspended				= 3,
	FSM_LoggingPosition			= 4
};

// 3. Count your states
#define FSM_States_Count		5

// 4. Define events used in your state machine
//    Leave 'Timeout' at the beginning
enum FSM_Events {
	GPSFSM_Timeout					= 0,
	GPSFSM_GotFix					= 1,
	GPSFSM_GotRMC					= 2,
	GPSFSM_Suspend					= 3
};

// 5. Count your events
#define FSM_Events_Count 	4


// 6. Define timeouts for your states
//    When no event happens within the timeout value, the state change will be executed (signalled as timeout)
//    0 stands for no timeout.
//    The order must correspond to the order in your states definition
uint32_t fsm_timeouts[FSM_States_Count] = {
	TICKS_5MINUTES,
	TICKS_MINUTE,
	TICKS_5SECONDS, 			
	TICKS_10MINUTES,
	TICKS_10SECONDS
};

// 7. Define your state transitions.
//    Horiziontal:	current state
//    Vertical:		The event
//    Entry:		Ation to call and the next state.
 FSM_Entry fsm_transitions[FSM_Events_Count][FSM_States_Count] = {

/*						Idle								WaitingForFix						WaitingForRMC 						Suspend							Logging Position*/
/* Timeout  */ 		{	{ EnableGPS, FSM_WaitingForFix }, 	{ DisableGPS, FSM_Idle }, 			{ DisableGPS, FSM_Idle }, 			{ NoAction, FSM_Idle },			{ NoAction, FSM_WaitingForRMC }  },
/* GotFix */ 		{	{ NoAction, FSM_WaitingForRMC }, 	{ NoAction, FSM_LoggingPosition }, 	{ NULL, 0 }, 						{ NULL, 0 },					{ NULL, 0 }  },
/* GotRMC */		{	{ NULL, 0 }, 						{ NULL, 0 }, 						{ DisableGPS, FSM_Idle }, 			{ NULL, 0 },					{ NULL, 0 }  },
/* Suspend */ 		{	{ DisableGPS, FSM_Suspended }, 		{ DisableGPS, FSM_Suspended }, 		{ DisableGPS, FSM_Suspended }, 		{ NoAction, FSM_Suspended },	{ DisableGPS, FSM_Suspended }  }
};
 
// 8. Initialize the Finite State Machine (FSM)
FSM_NEW(fsm_tracking, FSM_States_Count, FSM_Events_Count, fsm_transitions, fsm_timeouts);


// 9. Implement the init function and your actions

void Handler_GPS_GGA(const char* nmea, const logentry_t* entry) {
	// log every gga received
	LogEngine_Save(entry);
	
	// increase the fix counter
	FSM_Tracking_FixCounter++;
	
	// cache this GGA data for other modules
	memcpy(&FSM_Trackin_LastGGA, entry, sizeof(logentry_t));
	
	// trigger FSM state transition
	FSM_ProcessEvent(&fsm_tracking, GPSFSM_GotFix);
}

void Handler_RMC(const char* nmea, const time_t* time, const rawtime_t* rawtime) {
	// the GPS receiver is updating time and logging it, no action is required
	
	// trigger FSM state transition
	FSM_ProcessEvent(&fsm_tracking, GPSFSM_GotRMC);
}

/*void Handler_SCP() {
	uint32_t pressure;
	uint16_t temperature;
	rawtime_t rt;
	
    if (SCP_Read(NULL, NULL, &temperature, &pressure, &rt)) {
    	logentry_t scp;
    	LogEngine_PrepareLogEntry(&scp);
    	scp.TypeId = SCP;
    	
    	scp.Payload.SCP.Pressure = pressure;
    	scp.Payload.SCP.Temperature = temperature;
    	scp.Timestamp = rt;
    	
    	LogEngine_Save(&scp);
    }
}*/


static void _UpdateConfiguration() {
	if (false == AppConfiguration.Tracking_Configured) {
		AppConfiguration.Tracking_IdleTime 		= DEFAULT_TRACKING_IDLE_TIME;
		AppConfiguration.Tracking_MaxTTFTime 	= DEFAULT_TRACKING_MAX_TTF_TIME;
		AppConfiguration.Tracking_LoggingTime 	= DEFAULT_TRACKING_LOGGING_TIME;
		AppConfiguration.Tracking_PowerUpDelay 	= DEFAULT_TRACKING_POWERUP_DELAY;
		
		AppConfiguration.Tracking_Configured 	= true; // since default values were loaded
	}
	
	if (false == AppConfiguration.Suppress_Configured) {
		AppConfiguration.Suppress_Activation	= DEFAULT_SUPPRESS_ACTIVATION;
		AppConfiguration.Suppress_Deactivation 	= DEFAULT_SUPPRESS_DEACTIVATION;
		
		AppConfiguration.Suppress_Configured 	= true; // since the default values were loaded
	}
	
	fsm_timeouts[FSM_Idle] 				= AppConfiguration.Tracking_IdleTime;
	fsm_timeouts[FSM_WaitingForFix] 	= AppConfiguration.Tracking_MaxTTFTime;
	fsm_timeouts[FSM_LoggingPosition]	= AppConfiguration.Tracking_LoggingTime;
}

void FSM_Tracking_Init() {
	// process the configuration
	_UpdateConfiguration();
	
	// use this chance to measure the voltage of the battery
	FSM_Tracking_LastVoltage = ADC_Devices_Battery_Read(NULL, NULL);

	// register handlers for GGA and RMC
	Handler_GPS_NMEA_GGA = Handler_GPS_GGA;
	Handler_GPS_NMEA_RMC = Handler_RMC;
	
	// NoSCP mode
	//SCP_Enable(false);
	//Handler_SCP_Results = Handler_SCP;
	
	FSM_Tracking_FixCounter = 0;

	/**
	 * Enable the GPS:
	 * 
	 * This is new requirement defined by users of the device.
	 * 
	 * Not using the EnableGPS function because it is processing the 'schedule' config
	 * and now it's not required.
	 */
	GPS_PowerOn();
	GPS_StartProcessingNMEA();
	
	
	fsm_startparameters_t init;
	init.FSM			= &fsm_tracking;
	init.Delay			= AppConfiguration.Tracking_PowerUpDelay;
	init.InitialState	= FSM_Idle;
	
	FSM_Start(&init);
}

void FSM_Tracking_Abort() {
	FSM_Stop(&fsm_tracking);
	DisableGPS();
}

void FSM_Tracking_Suspend() {
	FSM_ProcessEvent(&fsm_tracking, GPSFSM_Suspend);
}

void EnableGPS() {
	// NoSCP mode
	//SCP_TriggerMeasurement();
	
	bool gpsActivationSuppressed = false;
	
	// evaluate the suppress configuration
	if (AppConfiguration.Suppress_Configured) {
		
		// get the current time
		rawtime_t currentRawtime;
		RTC_GetRawTime(&currentRawtime);
		
		// and continue only if the current time is valid
		if (RTC_ValidTime(&currentRawtime)) {
			time_t currentTime;
			RTC_ConvertToTime(&currentRawtime, &currentTime);
			
			// now evaluate the suppress configuration
			if (AppConfiguration.Suppress_Deactivation >=	 AppConfiguration.Suppress_Activation) {
				gpsActivationSuppressed = currentTime.hour >= AppConfiguration.Suppress_Activation && currentTime.hour < AppConfiguration.Suppress_Deactivation; 
			} else {
				gpsActivationSuppressed = currentTime.hour >= AppConfiguration.Suppress_Activation || currentTime.hour < AppConfiguration.Suppress_Deactivation;
			}
		}
	}
	
	// activate gps only if this was allowed
	if (gpsActivationSuppressed == false) {
		GPS_PowerOn();
		GPS_StartProcessingNMEA();
	} else {
		LOGDBG("EnableGPS suppressed.");
	}
}

#if (THRESHOLD_BATTERY_PROTECTION)
	void _BatteryCritical(uint16_t notUsed) {
		GPS_PowerOn();
		FSM_Tracking_LastVoltage = ADC_Devices_Battery_Read(NULL, NULL);
		GPS_PowerOff();
	
		// if the votage is smaller than the critical value again..
		float fbattery;
		ADC_Devices_Battery_Convert(FSM_Tracking_LastVoltage, &fbattery, NULL);
		
		if (fbattery < THRESHOLD_BATTERY_PROTECTION_LEVEL) {
			Network_PowerOff();
	
			LogEngine_SaveText("BattProtect");
	
			MATAKI_CriticalLPM4(5);
		}
	}
#endif

void DisableGPS() {
	// NoSCP mode
	//SCP_TriggerMeasurement();

	// perform battery measurement but only, if GPS is enabled
	FSM_Tracking_LastVoltage = ADC_Devices_Battery_Read(NULL, NULL);
	
	// disable the Venus chip
	GPS_PowerOff();

	// Save the recent voltage information
	if (FSM_Tracking_LastVoltage > 0) {
		logentry_t entry;
		LogEngine_PrepareLogEntry(&entry);
		entry.TypeId = Battery;
		entry.Payload.Battery.Value = FSM_Tracking_LastVoltage;
		
		LogEngine_Save(&entry);

#if (THRESHOLD_BATTERY_PROTECTION)
		// if the votage is smaller than the critical value...
		float fbattery;
		ADC_Devices_Battery_Convert(FSM_Tracking_LastVoltage, &fbattery, NULL);
		
		if (fbattery < THRESHOLD_BATTERY_PROTECTION_LEVEL) {
			LOGERR("LowBattWarning");
			Timers_Add(_BatteryCritical, TICKS_10SECONDS, 0);
		}
#endif
		
	}
}

COMMAND(setgps, "Configures the GPS experiment", args) {
	if (args->ArgumentCount != 4) {
		printf("Help: setgps [power up delay] [sleep time between fixes] [max time to wait for fix] [logging time after a fix]\r\n"
			   "Remark: 0 stands for infinite\r\n");

		printf("Power up delay:\t%lus\r\n"
			   "Sleep time:\t%lus\r\n"
			   "Max. TTF:\t%lus\r\n"
			   "Logging:\t%lus\r\n",
			   
			   AppConfiguration.Tracking_PowerUpDelay 	/ TICKS_SECOND,
			   AppConfiguration.Tracking_IdleTime 		/ TICKS_SECOND,
			   AppConfiguration.Tracking_MaxTTFTime 	/ TICKS_SECOND,
			   AppConfiguration.Tracking_LoggingTime 	/ TICKS_SECOND
			   
			   );
		
		snprintf(args->ResponseString, args->ResponseStringMaxLength,
				
				"Current configuration: {%lu,%lu,%lu,%lu}",
				
				AppConfiguration.Tracking_PowerUpDelay 	/ TICKS_SECOND,
				AppConfiguration.Tracking_IdleTime 		/ TICKS_SECOND,
				AppConfiguration.Tracking_MaxTTFTime 	/ TICKS_SECOND,
				AppConfiguration.Tracking_LoggingTime 	/ TICKS_SECOND
				
				);
		
		
		return;
	}
	
	// parameters count is ok, update the configuration
	AppConfiguration.Tracking_PowerUpDelay		= (uint32_t) atol(args->ArgumentArray[0]) * TICKS_SECOND;
	AppConfiguration.Tracking_IdleTime			= (uint32_t) atol(args->ArgumentArray[1]) * TICKS_SECOND;
	AppConfiguration.Tracking_MaxTTFTime		= (uint32_t) atol(args->ArgumentArray[2]) * TICKS_SECOND;
	AppConfiguration.Tracking_LoggingTime		= (uint32_t) atol(args->ArgumentArray[3]) * TICKS_SECOND;
	
	AppConfiguration.Tracking_Configured		= true;
	
	Configuration_Save();
	
	snprintf(args->ResponseString, args->ResponseStringMaxLength,
			
			"{%lu,%lu,%lu,%lu} Changes saved, reset to activate.",
			
			AppConfiguration.Tracking_PowerUpDelay 	/ TICKS_SECOND,
			AppConfiguration.Tracking_IdleTime 		/ TICKS_SECOND,
			AppConfiguration.Tracking_MaxTTFTime 	/ TICKS_SECOND,
			AppConfiguration.Tracking_LoggingTime 	/ TICKS_SECOND
			
			);
}

COMMAND(suppress, "Configures down time for the GPS (overrides setgps)", args) {
	if (args->ArgumentCount != 2) {
		printf("Help: suppress [HH of suppress activation] [HH of suppress deactivation]\r\n"
			   "Remark:\t suppress 10 13 means the GPS will be disabled when time >= 10 and time < 13\r\n"
			   "\t resulting in GPS activity from 10:00:00 till 12:59:59\r\n"
				"\t suppress has no effect when the time isn't correct\r\n"
				"\t suppress 10 10 disables this feature (when HH are equal the condition above is always false\r\n");

		if (AppConfiguration.Suppress_Configured) {
			snprintf(args->ResponseString, args->ResponseStringMaxLength,
					
					"Current configuration: {%u,%u}",
					
					AppConfiguration.Suppress_Activation,
					AppConfiguration.Suppress_Deactivation);
		}
		
		return;
	}
	
	AppConfiguration.Suppress_Activation	= atoi(args->ArgumentArray[0]);
	AppConfiguration.Suppress_Deactivation 	= atoi(args->ArgumentArray[1]);
	AppConfiguration.Suppress_Configured 	= true;
	
	Configuration_Save();
	
	snprintf(args->ResponseString, args->ResponseStringMaxLength,
			
			"{%u,%u} Changes saved, reset to activate.",
			
			AppConfiguration.Suppress_Activation,
			AppConfiguration.Suppress_Deactivation
			
			);
}

#ifndef DOXYGEN_PUBLIC_DOC
/**
* Copyright (c) Microsoft Corporation. All rights reserved.
* Licensed under the Apache License, Version 2.0 (the "License"); you may not use these files except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
* Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.
 
* Copyright (c) 2012 Freie Universitaet Berlin, Tomasz Naumowicz. All rights reserved.
* 
* Redistribution and use in source and binary forms, with or without modification, are
* permitted provided that the following conditions are met:
* 
* - Redistributions of source code must retain the above copyright notice, this list of
* conditions and the following disclaimer.
* 
* - Redistributions in binary form must reproduce the above copyright notice, this list
* of conditions and the following disclaimer in the documentation and/or other materials
* provided with the distribution.
* 
* - Neither the name of the Freie Universitaet Berlin nor the
* names of its contributors may be used to endorse or promote products
* derived from this software without specific prior written permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Freie Universitaet Berlin BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#endif // DOXYGEN
