/**
 * @file		FSM_Tracking.h
 * @brief		State machine controling the tracking features.
  * The \ref FSM.h contains the implementation of the Finite State Machine.
 * This file contains the description of the project specific FSM.
 * 
 * The Tracking Finite State Machine controls the hevahior of the GPS receiver, battery measurements and SCP
 * (temperature and pressure) measurements.
 */

#ifndef FSM_TRACKING_H_
#define FSM_TRACKING_H_

#include "Application.API.h"

/**
 * Starts the Tracking state machine.
 * 
 * The behavior of the state machine:
 * 
 * @dot
 * digraph fsm {
 * 		graph 					[fontname=Helvetica, fontsize=10];
 * 		edge 					[fontname=Helvetica, fontsize=10];
 *
 *  
 *		node 					[shape=doublecircle, fontname=Helvetica, fontsize=10];
 * 		initwait				[label = "Initial Delay\nGPS Enabled"];
 * 
 *		node 					[shape=circle, fontname=Helvetica, fontsize=10];
 * 		idle 					[label = "Idle\nGPS Disabled", color="darkgreen", style="filled", fillcolor="lightgrey"];
 *		waiting 				[label = "Waiting for a Fix\nGPS Enabled"];
 * 		logging 				[label = "Logging Position\nGPS Enabled"]
 * 		waitingrmc 				[label = "Waiting for RMC\nGPS Enabled"];
 * 		stop					[label = "Stopped\nGPS Enabled"];
 *
 * 		initwait -> waiting		[label="Configurable Timeout\n(Start Tracking Mode)", style="dashed", color="darkgreen", fontcolor="blue"];
 * 		initwait -> stop		[label="'abort'"];
 * 		initwait -> waitingrmc	[label="Got a Fix"];
 * 		idle -> waiting 		[label="Configurable Timeout\n(Enable GPS, SCP)", style="dashed", weight="8", color="darkgreen", fontcolor="blue"];
 * 		waiting -> idle 		[label="Configurable Timeout\n(Disable GPS, SCP)", style="dashed", weight="8", fontcolor="blue"];
 * 		waiting -> logging 		[label="Got a Fix", weight="8", color="darkgreen"];
 * 		logging -> waitingrmc 	[label="Configurable Timeout", style="dashed", weight="8", color="darkgreen", fontcolor="blue"];
 * 		waitingrmc -> idle 		[label="Got RMC\n(Disable GPS, SCP)", weight="8", color="darkgreen"];
 * 		waitingrmc -> idle 		[label="Short Timeout\n(Disable GPS, SCP)", style="dashed"];
 * 
 *  
 *		node 					[shape=circle, fontname=Helvetica, fontsize=10, color="gray", fontcolor="gray"];
 * 		suspend 				[label = "GPS Disabled"];
 * 
 * 		edge 					[fontname=Helvetica, fontsize=10, color="gray", fontcolor="gray"];
 * 		initwait -> suspend 	[label="Suspend GPS\n(Disable GPS, SCP)"];
 * 		waiting -> suspend 		[label="Suspend GPS\n(Disable GPS, SCP)"];
 * 		idle -> suspend 		[label="Suspend GPS\n(Disable GPS, SCP)"];
 * 		logging -> suspend 		[label="Suspend GPS\n(Disable GPS, SCP)"];
 * 		waitingrmc -> suspend	[label="Suspend GPS\n(Disable GPS, SCP)"];
 * 		stop -> suspend			[label="Suspend GPS\n(Disable GPS, SCP)"];
 * 		suspend -> idle 		[label="10 Minutes Timeout\n(Disable GPS, SCP)", style="dashed"];
 * 		suspend -> suspend 		[label="SuspendGPS"];
 * 
 * 
 * }
 * @enddot
 */
void FSM_Tracking_Init();

void FSM_Tracking_Abort();

logentry_t	FSM_Trackin_LastGGA;
uint16_t	FSM_Tracking_LastVoltage;
uint16_t	FSM_Tracking_FixCounter;

void FSM_Tracking_Suspend();

#endif /*FSM_GPS_H_*/

#ifndef DOXYGEN_PUBLIC_DOC
/**
* Copyright (c) Microsoft Corporation. All rights reserved.
* Licensed under the Apache License, Version 2.0 (the "License"); you may not use these files except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
* Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.
 
* Copyright (c) 2012 Freie Universitaet Berlin, Tomasz Naumowicz. All rights reserved.
* 
* Redistribution and use in source and binary forms, with or without modification, are
* permitted provided that the following conditions are met:
* 
* - Redistributions of source code must retain the above copyright notice, this list of
* conditions and the following disclaimer.
* 
* - Redistributions in binary form must reproduce the above copyright notice, this list
* of conditions and the following disclaimer in the documentation and/or other materials
* provided with the distribution.
* 
* - Neither the name of the Freie Universitaet Berlin nor the
* names of its contributors may be used to endorse or promote products
* derived from this software without specific prior written permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Freie Universitaet Berlin BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#endif // DOXYGEN
