/**
 * @file
 * @ingroup		DriverAPI
 * @brief		@b Driver: CMA Accelerometer sensor
 * 
 * The CMA is a three-axis accelerometer developed by VTI (http://www.vti.fi/en/support/).
 * 
 * This driver enables access to accelerometer data.
 * 
 */

#ifndef CMA3000_API_H_
#define CMA3000_API_H_

#include <stdint.h>
#include "../System/Types.h"
#include "../System/RTC.API.h"



/**
 * @brief @b Handler: Definition of the handler for processing of results of a accelerometer measurement
 */
//typedef void(*fp_cma_data_handler_t) (void);

/**
 * @ingroup	Handlers

 * @brief @b Handler: Access to results of an accelerometer measurement (see \ref Handlers for more handlers).
 * 
 * You need to process notifications about available results and load them from the sensor.
 * Any subsequent requests for a measurement will fail if previous results aren't loaded first.
 *  
 * Example:
 * @code
 * // sample data handler
 * void MyHandler() {
 *	float temperature;
 *	uint32_t pressure;
 *	uint16_t rawTemperature;
 *	uint32_t rawPressure;
 *	rawtime_t rt;
 *	time_t t;
 * 
 *	if (SCP_Read(&temperature, &pressure, &rawTemperature, &rawPressure, &rt)) {
 *		// here you can access the data and log it or transmit it over the radio
 *	}
 * }
 * 	
 * // somewhere in your application code, e.g. in the Application_Init() function
 * // ..
 * Handler_SCP_Results = MyHandler;
 * // ...
 * @endcode
 * 
 * @remarks		Only one handler can be registered.
 */
//fp_cma_data_handler_t Handler_CMA_Results;


/**
 * @name		Managing the CMA device
 * 
 * @{
 */
 
/**
 * @brief 	Enables the CMA device
 * 
 * Enables the CMA sensor: brings it back from the power down mode.
 * 
 * @returns true on success and false on failure
 */

bool CMA_Enable(bool printProgress);

/**
 * @brief	Disables the CMA device
 * 
 * Puts the CMA sensor into the power down mode.
 */
void CMA_Disable();
void CMA_DisableI2C();

/** @} */

/**
 * @name	Performing measurements
 * 
 * @{
 */

/**
 * @brief 	Triggers the accelerometer measurement and returns.
 * 
 * The measurement is executed in background by the CMA sensor.
 * The results can be read later using the \ref CMA_Read function.
 * 
 * @remarks To driver developers: the application will be notified by a ISR registered on the DRDY pin.
 * This ISR should also store the timestamp of the measurement.
 */
//bool CMA_TriggerMeasurement();

/**
 * describe the measurement mode and assoiate settings (see similar for the pressure sensing code)
 */
//bool CMA_MeasureBlocking(uint8_t* XOut, uint8_t* YOut, uint8_t* ZOut);

/** @} */



/**
 * @name	Working with measurement results
 * 
 * @{
 */

/**
 * @brief	Reads results from a recent accelerometer measurement.
 *
 * Describe results, perhaps similar to the SCP code...
 * 
 * Example:
 * @code
 * // sample data handler
 * void MyHandler() {
 *	float temperature;
 *	uint32_t pressure;
 *	uint16_t rawTemperature;
 *	uint32_t rawPressure;
 *	rawtime_t rt;
 *	time_t t;
 * 
 *	if (SCP_Read(&temperature, &pressure, &rawTemperature, &rawPressure, &rt)) {
 *		// here you can access the data and log it or transmit it over the radio
 *	}
 * }
 * 	
 * // somewhere in your application code, e.g. in the Application_Init() function
 * // ..
 * Handler_SCP_Results = MyHandler;
 * // ...
 * @endcode
 *  
 * @param[out]	temperatureOutOptional		Result of the measurement, converted to floating point type. Set to NULL if that type is not required.
 * @param[out]	pressureOutOptional			Result of the measurement, converted. Set to NULL if that type is not required.
 * @param[out]	rawTemperatureOutOptional	Result of the measurement, converted to floating point type. Set to NULL if that type is not required.
 * @param[out]	rawPressureOutOptional		Result of the measurement in internal raw format. Set to NULL if that type is not required.
 * @param[out]	rawPressureOutOptional		Result of the measurement in internal raw format. Set to NULL if that type is not required.
 * @param[out]	timestampOutOptional		Timestamp of the measurement. Set to NULL if that type is not required.
 * 
 * @returns true on success, false on failure
 */
bool CMA_Read(uint8_t* XOut, uint8_t* YOut, uint8_t* ZOut, rawtime_t* timestampOutOptional);
uint8_t CMA_ReadX();
uint8_t CMA_ReadY();
uint8_t CMA_ReadZ();

bool CMA_SetMode(uint16_t Acc_SamplingRate, uint16_t Acc_gRange);
/**
 * @brief		Converts results of the pressure and temperature measurement
 * 
 * Conversion should only be processed when the data needs to be displayed to the user. In most cases, e.g. logging, the 
 * raw pressure and raw temperature data should be processed.
 * 
 * @param[out]	temperatureOutOptional	Result of the conversion as a floating point type. Set to NULL if that type is not required.
 * @param[out]	pressureOutOptional		Result of the conversion. Set to NULL if that type is not required.
 * @param[in]	rawTemperatureIn		Result of the measurement provided by the \ref SCP_Read function.
 * @param[in]	rawPressureIn			Result of the measurement provided by the \ref SCP_Read function.
 */
void CMA_Convert(uint8_t* XOutOptional, uint8_t* YOutOptional, uint8_t* ZOutOptional, uint8_t XOutRaw, uint8_t YOutRaw, uint8_t ZOutRaw);

/** @} */

#endif /*CMA3000_API_H_*/

#ifndef DOXYGEN_PUBLIC_DOC
/**
* Copyright (c) Microsoft Corporation. All rights reserved.
* Licensed under the Apache License, Version 2.0 (the "License"); you may not use these files except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
* Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.
*/
#endif // DOXYGEN
