/**
 * @file		CMA3000.c
 * @ingroup		Driver
 * @brief		@b Driver: CMA Accelerometer sensor
 */
#include <msp430f2618.h>
#include <string.h>

#include "Device.h"

#include "../System/BitOperations.h"

#include "../System/USCI.API.h"
#include "../System/RTC.API.h"
#include "../System/Logging.h"
#include "../System/Timers.API.h"

#include "CMA3000.h"

/**
 * List of CMA3000 direct and indirect access registers and EEPROM registers.
 */
#define WHO_AM_I	0x00 //Identification register
#define REVID		0x01 //ASIC revision ID, fixed in metal
#define CTRL		0x02 //Configuration (por, operation modes)
#define STATUS		0x03 //Status (por, EEPROM parity)
#define RSTR		0x04 //Reset register
#define INT_STATUS	0x05 //Interrupt status register
#define DOUTX		0x06 //X channel output data register
#define DOUTY		0x07 //Y channel output data register
#define DOUTZ		0x08 //Z channel output data register
#define MDTHR		0x09 //Motion detection threshold value register
#define MDFFTMR		0x0A //Free fall and motion detection time register
#define FFTHR		0x0B //Free fall threshold value register
//0x0D-0x19 are reserved

/**
 * Power down mode (to be written to the CTRL register)
 */
#define M_POWERDOWN_DEF		0X31
#define M_POWERDOWN			0X3F

/**
 * Measurement modes (to be written to the CTRL register)
 */
#define MM_MEDSPEED_8g		0X33		//100 Hz
#define MM_HISPEED_8g		0X35		//400 Hz
#define	MM_LOSPEED_8g		0X37		//40 Hz

#define MM_MEDSPEED_2g		0XB3		//100 Hz
#define MM_HISPEED_2g		0XB5		//400 Hz
//#define	MM_LOWSPEED_2g		0X37		//40 Hz		//this mode is not supported by the component

/**
 * Detection modes (to be written to the CTRL register)
 */
#define MDM_MOTIONDET		0x39
#define FFDM_LOWSPEED		0x3B
#define FFDM_HIGHSPEED		0x3D

/**
 * SPI, I2C, and Interrupt-related modes (to be written to the CTRL register)
 */
#define M_NOI2C				0x10
#define M_NoI2CNOINT		0x11

rawtime_t CMA_MeasurementTimestamp;

// Function takes in the 8-bit register address and 8-bit register value. 
void CMA3000_Write(uint8_t address, uint8_t data) { 
	address = address << 2;  
	address |= 0x02;    // Set write bit
 
	CLEAR( CMA_CS_OUT, CMA_CS_PIN );
	USCI_B0_SPI_Lock(0x1F, 0x00);

	USCI0_SPI_Comm(address);		// send read who_am_i
	USCI0_SPI_Comm(data);
	
	SET( CMA_CS_OUT, CMA_CS_PIN );
	USCI_B0_SPI_Release();

	return; 
} 

// Function takes in the 8-bit register address and 8-bit register value. 
uint8_t CMA3000_Read(uint8_t address) { 
	address = address << 2;  
 
	CLEAR( CMA_CS_OUT, CMA_CS_PIN );
	USCI_B0_SPI_Lock(0x1F, 0x00);

	USCI0_SPI_Comm(address);
	uint8_t response = USCI0_SPI_Comm(0x00);
	
	SET( CMA_CS_OUT, CMA_CS_PIN );
	USCI_B0_SPI_Release();

	return response; 
} 

bool CMA_Read(uint8_t* XOut, uint8_t* YOut, uint8_t* ZOut, rawtime_t* timestampOutOptional){
	// store the current timestamp
	RTC_GetRawTime(&CMA_MeasurementTimestamp);
	// publish the timestamp
	if (timestampOutOptional != NULL) {
		memcpy(
			timestampOutOptional,
			&CMA_MeasurementTimestamp,
			sizeof(rawtime_t)
		);
	}
	*XOut = CMA3000_Read(DOUTX);
	*YOut = CMA3000_Read(DOUTY);
	*ZOut = CMA3000_Read(DOUTZ);

	return true;
}
uint8_t CMA_ReadX(){
	return CMA3000_Read(DOUTX);
}
uint8_t CMA_ReadY(){
	return CMA3000_Read(DOUTY);
}
uint8_t CMA_ReadZ(){
	return CMA3000_Read(DOUTZ);
}

void CMA_Convert(uint8_t* XOutOptional, uint8_t* YOutOptional, uint8_t* ZOutOptional, uint8_t XOutRaw, uint8_t YOutRaw, uint8_t ZOutRaw){
	//convert the readings
	/*if (XOutOptional != NULL) {
		bool isNegative = (rawTemperatureIn & BITD) == BITD;
		if (isNegative) {*/
			/**
			 * the temperature was negative, special treatment is required.
			 * See scp1000_product_family_specification_rev_0.08-3.pdf, 2.2.3.1 Examples of temperature conversion to [C] for details
			 */
			
	/*		rawTemperatureIn = ~rawTemperatureIn;				// invert bits
			rawTemperatureIn = rawTemperatureIn & 0x3FFF; 		// mask bits [13:0]
			rawTemperatureIn++;									// add 1
		}
		
		*temperatureOutOptional = rawTemperatureIn;
		*temperatureOutOptional = *temperatureOutOptional / 20;
		
		if (isNegative) {
			*temperatureOutOptional =  0.0f - *temperatureOutOptional;
		}
	}

	if (pressureOutOptional != NULL) {
		uint32_t rawPressureInCopy = rawPressureIn;
		rawPressureInCopy = rawPressureInCopy >> 2;
		*pressureOutOptional = rawPressureInCopy;
	}
	*XOutOptional = XOutRaw;
	*YOutOptional = YOutRaw;
	*ZOutOptional = ZOutRaw;*/
}

void CMA_DisableI2C() {
	/*uint8_t i2c_disabled = CMA3000_Read(0x02);	// CMA3000 confirm i2c off
	LOGDBG("CMA3000: %u", i2c_disabled);
	CMA3000_Write(0x02, 0x11);					// CMA3000 i2c off
	Timers_Block(1024);
	i2c_disabled = CMA3000_Read(0x02);	// CMA3000 confirm i2c off
	LOGDBG("CMA3000: %u", i2c_disabled);
	Timers_Block(1024);
	
	i2c_disabled = CMA3000_Read(0x02);	// CMA3000 confirm i2c off
	LOGDBG("CMA3000: %u", i2c_disabled);
	i2c_disabled &= 0x10;
	LOGDBG("CMA3000: %u", i2c_disabled);
	if (i2c_disabled > 0)
		LOGDBG("CMA3000: I2C is disabled");
	else
		LOGDBG("CMA3000: error disabling I2C!");
	//CMA3000_Write(0x02, 0x1F);					// CMA3000 power down
	*/
	
	CMA3000_Write(CTRL, M_NoI2CNOINT);				// CMA3000 i2c off (and interrupt disabled)
	
	//this is perhaps not necessary here, to be tested
	Timers_Block(30);
	
	uint8_t i2c_disabled = CMA3000_Read(CTRL);		// CMA3000 confirm i2c off
	i2c_disabled &= M_NoI2CNOINT;
	if (i2c_disabled > 0)
		LOGDBG("CMA3000: I2C disabled");
	else{
		LOGDBG("CMA3000: error disabling I2C!");
	}
}

void CMA_Disable() {
	CMA3000_Write(CTRL, M_POWERDOWN);				// CMA3000 force to power down
	Timers_Block(30);
	uint8_t cma_disabled = CMA3000_Read(CTRL);		// CMA3000 confirm power down
	cma_disabled &= M_POWERDOWN;
	if (cma_disabled > 0)
		LOGDBG("CMA3000: disabled");
	else{
		LOGDBG("CMA3000: error disabling!");
	}
}

bool CMA_Enable(bool printProgress) {
	LOGDBG("Enabling CMA3000...");
	//CMA_DisableI2C();
	CMA3000_Write(CTRL, MM_MEDSPEED_2g);			// CMA3000 measaurement mode enabled
	Timers_Block(30);
	uint8_t cma_enabled = CMA3000_Read(CTRL);		// CMA3000 confirm mode
	cma_enabled &= MM_MEDSPEED_2g;
	if (cma_enabled > 0){
		LOGDBG("CMA3000: 2g medium speed mode enabled");
		return true;
	}
	else{
		LOGDBG("CMA3000: error enabling!");
		return false;
	}
	return true;
}

/*
bool CMA_TriggerMeasurement() {
	// check the preconditions:
	
	// 1. check whether the device is active
	if (CMA_PD_POUT & CMA_PD_PIN) {
		// the device is in the power down mode, it wasn't activated or the activation has failed
		LOGDBG("CMAt!PD");

		return false;
	}

	// 2. the request can only be issued when the DRDY line is low
	if (CMA_DRDY_PINPUT & CMA_DRDY_PIN) {
		LOGDBG("CMAt!DRDY");

		return false; // a dummy read could be executed here but it's actually user's resposibility
	}
	
	// the measurement is triggered by the low-to-high edge of the TRIG pin
	// the TRIG is set to low during reading of the last measrurement, so it's safe to assume
	// the TRIG is low 
	
	// 3. the request can only be issued when the TRIG is low
	if (CMA_TRIG_POUT & CMA_TRIG_PIN) {
		LOGDBG("CMAt!TRIG");
		
		return false;
	}
	
	//
	// preconditions check passed
	//
	
	// store the current timestamp
	RTC_GetRawTime(&CMA_MeasurementTimestamp);

	// trigger the conversion
	SET(CMA_TRIG_POUT, CMA_TRIG_PIN);
	
	return true;
}
*/

bool CMA_SetMode(uint16_t Acc_SamplingRate, uint16_t Acc_gRange){
	uint8_t cma_enabled = 0;
	if((Acc_SamplingRate == 40) && (Acc_gRange == 8)) {
		CMA3000_Write(CTRL, MM_LOSPEED_8g);			// CMA3000 measaurement mode enabled
		Timers_Block(30);
		cma_enabled = CMA3000_Read(CTRL);		// CMA3000 confirm mode
		cma_enabled &= MM_LOSPEED_8g;
	}
	else if((Acc_SamplingRate == 100) && (Acc_gRange == 8)) {
		CMA3000_Write(CTRL, MM_MEDSPEED_8g);			// CMA3000 measaurement mode enabled
		Timers_Block(30);
		cma_enabled = CMA3000_Read(CTRL);		// CMA3000 confirm mode
		cma_enabled &= MM_MEDSPEED_8g;
	}
	else if((Acc_SamplingRate == 400) && (Acc_gRange == 8)) {
		CMA3000_Write(CTRL, MM_HISPEED_8g);			// CMA3000 measaurement mode enabled
		Timers_Block(30);
		cma_enabled = CMA3000_Read(CTRL);		// CMA3000 confirm mode
		cma_enabled &= MM_HISPEED_8g;
	}
	else if((Acc_SamplingRate == 100) && (Acc_gRange == 2)) {
		CMA3000_Write(CTRL, MM_MEDSPEED_2g);			// CMA3000 measaurement mode enabled
		Timers_Block(30);
		cma_enabled = CMA3000_Read(CTRL);		// CMA3000 confirm mode
		cma_enabled &= MM_MEDSPEED_2g;
	}
	else if((Acc_SamplingRate == 400) && (Acc_gRange == 2)) {
		CMA3000_Write(CTRL, MM_HISPEED_2g);			// CMA3000 measaurement mode enabled
		Timers_Block(30);
		cma_enabled = CMA3000_Read(CTRL);		// CMA3000 confirm mode
		cma_enabled &= MM_HISPEED_2g;
	}
	if (cma_enabled > 0){
		LOGDBG("CMA3000: measurement mode set");
		return true;
	}
	else{
		LOGDBG("CMA3000: error setting mode!");
		return false;
	}
}

/**
* Copyright (c) Microsoft Corporation. All rights reserved.
* Licensed under the Apache License, Version 2.0 (the "License"); you may not use these files except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
* Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.
*/