/**
 * @file		GPS.API.h
 */

#include <string.h>
#include <stdlib.h>
#include <mspgcc/ringbuffer.h>

#include "Device.h"

#include "../System/System.h"
#include "../System/BitOperations.h"
#include "../System/Logging.h"
#include "../System/smallio.h"

#include "../System/RTC.API.h"
#include "../System/Timers.API.h"
#include "../System/USCI.API.h"
#include "../System/Commands.API.h"
#include "../System/Watchdog.API.h"

//#include "SCP1000-D01.h"
//#include "SCP1000-D01.API.h"

#include "NMEA.h"

#include "GPS.h"
#include "GPS.API.h"

/**
 * Maximum size of of the buffer used for communication with the GPS chip (replies to commands are being cached in this buffer)
 */ 
#define		GPS_RX_BUFFER_MAX_LENGTH			64

#ifndef		GPS_RMC_UPDATE_RATE
/*
 * The RMC sentence contains date and time information received from the GPS system.
 * The frequency with which the RMC sentece is provicdd by the GPS receiver depends on the system position rate.
 * The number provided here specifies the requency relative to the system position rate.
 * 
 * Example:
 * 	Setting the value to 5 means, the RMC sentence will be delivered every 5th time. Depending on the system position rate, e.g. when
 * 	the system position rate is set to 1Hz, the RMC sentence will be delivered every 5 seconds
 * 
 *  Supported values for the RMC update rate: 0, 1, 2, 3, 4, 5
 */
	#define		GPS_RMC_UPDATE_RATE		"5"
#endif

/**
 * Ringbuffer used to cache data received from the GPS chip when it's capturing replies to command requests
 */
RINGBUFFER_NEW(_gps_rx_ringbuffer, GPS_RX_BUFFER_MAX_LENGTH);

/**
 * Internal counter used to count number of valid binary messages
 * received from the GPS chip and cached in the ring buffer.
 * this value is increased by the ISR and reduced by the read message function.
 */
volatile uint8_t _command_replies_in_buffer;

/// states used by the state machines while reading and receiving binarry messages
enum gps_message_processing_states {
	WaitForSOS_1,		// $
	WaitForSOS_2,		// P
	WaitForSOS_3,		// M
	WaitForSOS_4,		// T
	WaitForSOS_5,		// K
	WaitForEOS_1,		// \r
	WaitForEOS_2		// \n
};

/// this variable is used to maintain the state of the state machine while it's receiving messages via ISR byte by byte
volatile enum gps_message_processing_states _gps_message_receive_state;

/// command response parts
#define		SOS_1		'$'
#define		SOS_2		'P'
#define		SOS_3		'M'
#define		SOS_4		'T'
#define		SOS_5		'K'
#define		EOS_1		'\r'
#define		EOS_2		'\n'

/**
 * 0 NMEA_SEN_GLL, // GPGLL interval - Geographic Position - Latitude longitude
 * 1 NMEA_SEN_RMC, // GPRMC interval - Recomended Minimum Specific GNSS Sentence
 * 2 NMEA_SEN_VTG, // GPVTG interval - Course Over Ground and Ground Speed
 * 3 NMEA_SEN_GGA, // GPGGA interval - GPS Fix Data
 * 4 NMEA_SEN_GSA, // GPGSA interval - GNSS DOPS and Active Satellites
 * 5 NMEA_SEN_GSV, // GPGSV interval - GNSS Satellites in View
 * 6 NMEA_SEN_GRS, // GPGRS interval - GNSS Range Residuals
 * 7 NMEA_SEN_GST, // GPGST interval - GNSS Pseudorange Errors Statistics
 * 
 *                                                                 0 1 2 3 4 5 9 7 8 9 A B C D E F 
 */
const char gps_command_Configure_NMEA_Messages[]			= "314,0," GPS_RMC_UPDATE_RATE ",0,1,0,0,0,0,0,0,0,0,0,0,0,0,0";
const char gps_command_Configure_NMEA_Messages_All[]		= "314,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0";

const char gps_command_Test[]								= "000";
const char gps_command_Set_NMEA_Baudrate_9600[]				= "251,9600";
const char gps_command_Set_NMEA_Baudrate_38400[]			= "251,38400";
const char gps_command_Set_NMEA_Baudrate_57600[]			= "251,57600";
const char gps_command_Set_NMEA_Baudrate_115200[]			= "251,115200";
const char gps_command_Position_Rate_1Hz[]					= "300,1000,0,0,0,0";
const char gps_command_Position_Rate_4Hz[]					= "300,250,0,0,0,0";
const char gps_command_Position_Rate_5Hz[]					= "300,200,0,0,0,0";
const char gps_command_Position_Rate_10Hz[]					= "300,100,0,0,0,0";
const char gps_command_Query_Position_Rate[]				= "400";

/**
 * @brief	Reads requested message from internal buffer.
 * 
 * All other messages stored in the internal buffer prior to the requested message that don't match the packet type will be discarded.
 * As concequence, if you request a unknown ID, all messages will be discarded.
 *  
 * @param[in]	packetType				PktType of the requested message.
 * @param[out]	bufferOut				Buffer which should be used to store the message.
 * @param[out]	buffer_length			The length of the bufferOut buffer. 
 * 
 * @returns		Number of bytes read. 0 on failure.
 */
uint8_t GPS_ReadCommandResponse(const char* packetType, uint8_t* bufferOut, uint8_t buffer_length) {
	// used to control the position in the output buffer, will be increased after every protocol conform character and set to 0 at failure
	uint8_t positionInBuffer = 0;
	
	// this is the main loop, it stops when there are no messages in the ringbuffer
	// or the out buffer is too small.
	while (_command_replies_in_buffer > 0 && positionInBuffer < buffer_length) {
		// read first message from the buffer
		while (positionInBuffer < buffer_length) {
			uint8_t byteFromBuffer = ringbuffer_get(&_gps_rx_ringbuffer);
			if (byteFromBuffer == -1) {
				// the buffer was empty, error. failure path
				_command_replies_in_buffer = 0; 
				return 0;
			}
			
			// now, copy the character into the output buffer
			bufferOut[positionInBuffer] = byteFromBuffer;
			
			// check whether the end of the message was found and break the inner while loop in this case
			if (byteFromBuffer == '\0') {
				_command_replies_in_buffer--;
				break;
			}

			// prepare the index for the next character (it might be, it will be reset in the state machine code!)
			positionInBuffer++;
		}
		
		// RF DEBUGGING - Print out buffer contents
        /*uint8_t idx;
	    for (idx = 0; idx < positionInBuffer; idx++) {
	        LOGDBG("%c", bufferOut[idx]);
	    }*/
        

		// now investigate if that message matches the requested packet type
		if (!strncmp( (char*) bufferOut, packetType, strlen(packetType) )) {
			// the prefix matches. accept the message. success path
			return positionInBuffer;
		} else {
			// the packet type doesn't match, retry
			positionInBuffer = 0;
		}
	} // while messagesinbuffer > 0
	
	// failure path:
	positionInBuffer = 0;
			
	return positionInBuffer;
}

/**
 * Blocks until a message with a specified packet type arrives.
 * A read is executed few times with a blocking delay in between.
 * 
 * @remarks		This function blocks for max. 50*265*122us = ca0.8 s
 * 
 * @returns		True on successful message delivery and False on failure or no reply
 */
bool GPS_BlockUntilPacketTypeReceived(const char* packetType, uint8_t* bufferOut, uint8_t buffer_length) {
	uint8_t retryCounter = 50;
	
	while (retryCounter > 0) {
		Timers_Block(256);

		uint8_t readLength = GPS_ReadCommandResponse(packetType, bufferOut, buffer_length);

		if (readLength > 0) {
		    LOGDBG("GPS: Got ACK from GPS, length: %u", readLength);
			return true;
		}

		// no ACK received, retry
		retryCounter--;
	}
	
	return 0;
}

/**
 * Submits a command to the GPS chip.
 * 
 * The functions expects the GPS to be enabled and UART A1 configured.
 */
void GPS_SubmitCommand(const char* command) {
	/**
	 * Command Format:
	 * 
	 * [const]					[ command parameter					]	[const]				[computed]				[const]
	 * Start Of Sequence (SOS)	PktType + Optional DataField			end of DataField	Checksum (1 byte)		End Of Sequence (EOS)
	 * "$PMTK" (5 bytes)		...										"*"					(as hex, two digits)	"\r\n"
	 */
	
	const uint8_t StartOfSequence[]	= "$PMTK";  
	const uint8_t EndOfSequence[]	= "\r\n";

	// compute the length of the command
	uint16_t cmd_length = strlen(command);
	
	//// now, the message will be transmitted
	
	// SOS
	USCI_A1_UART_TX_Buffer(5, StartOfSequence);

	// The command
	USCI_A1_UART_TX_Buffer(cmd_length, (uint8_t*) command);
	
	// end of DataField
	USCI_A1_UART_TX('*');
	
	// compute tge checksum of the data between "$" and "*"
	//  --> compute the checksum over "PMTK" and the command
	//		(the checksum ofer "PMTK" is already preset to 0x02)
	uint8_t cs = 0x02;
	for (; cmd_length > 0; cmd_length--) {
		cs ^= command[cmd_length - 1];
	}
	
	char cs_asString[3];
	snprintf(cs_asString, 3, "%02x", cs);
	
	// transmit the checksum
	USCI_A1_UART_TX_Buffer(2, (uint8_t*) cs_asString);
	
	
	// EOS
	USCI_A1_UART_TX_Buffer(2, EndOfSequence);
	
	// RF DEBUGGING - Print out buffer contents
    /*uint8_t idx;
    for (idx = 0; idx < cmd_length; idx++) {
        LOGDBG("%c", command[idx]);
    }*/

}

/**
 * This function intercepts characters received from the GPS receiver. It's only active during
 * command communication and processes only $PMTK sentences
 */
static bool _GPS_IncomingCommandResponseBytesHandler (uint8_t data) {
	if (ringbuffer_len(&_gps_rx_ringbuffer) == GPS_RX_BUFFER_MAX_LENGTH) {
		// critical state, the buffer is full, looks like the data isn't processed.. clean up.
		ringbuffer_clear(&_gps_rx_ringbuffer);
		_command_replies_in_buffer = 0;

		_gps_message_receive_state = WaitForSOS_1;
		LOGDBG("GPS COMMAND RESPONSE RINGBUFFER FULL!");
		
		SystemFlags.Flag.Pending_WakeUp_Request = true; // wake up the systen so that the error can be processed.
	}
	
	switch (_gps_message_receive_state) {
		case WaitForSOS_1 : {
			if (data == SOS_1) {
				_gps_message_receive_state = WaitForSOS_2;
			}

			break;
		}
		case WaitForSOS_2 : {
			if (data == SOS_2) {
				_gps_message_receive_state = WaitForSOS_3;
			} else {
				_gps_message_receive_state = WaitForSOS_1; // reset the state machine
			}

			break;
		}
		case WaitForSOS_3 : {
			if (data == SOS_3) {
				_gps_message_receive_state = WaitForSOS_4;
			} else {
				_gps_message_receive_state = WaitForSOS_1; // reset the state machine
			}

			break;
		}
		case WaitForSOS_4 : {
			if (data == SOS_4) {
				_gps_message_receive_state = WaitForSOS_5;
			} else {
				_gps_message_receive_state = WaitForSOS_1; // reset the state machine
			}

			break;
		}
		case WaitForSOS_5 : {
			if (data == SOS_5) {
				_gps_message_receive_state = WaitForEOS_1;
			} else {
				_gps_message_receive_state = WaitForSOS_1; // reset the state machine
			}

			break;
		}
		case WaitForEOS_1 : {
			if (data == EOS_1) {
				_gps_message_receive_state = WaitForEOS_2;
				break;
			}

			// the start sequence was detected, store everything until the end sequence starts
			ringbuffer_put(&_gps_rx_ringbuffer, data);
	
			break;
		}
		case WaitForEOS_2 : {
			if (data == EOS_2) {
				ringbuffer_put(&_gps_rx_ringbuffer, '\0');

				_command_replies_in_buffer++;
				
				_gps_message_receive_state = WaitForSOS_1;	
				
				SystemFlags.Flag.Pending_WakeUp_Request = true; // wake up the system so that the response can be processed
			}

			break;
		}
	}
	
	return true;
}

uint8_t fix_counter = 0;

/**
 * this handler intercepts NMEA sentences and forwards them to registered listeners.
 */
static void _GPS_NMEA_Processing_Handler(const char* input) {
	if (NMEA_ValidSentence(input) == false) {
		// try to find the end of string for debugging:
		uint8_t index;
		for (index = 0; index < 120; index++) {
			if (input[index] == 0x00)
				break;
		}

		LOGERR("CHKSUM FAIL (0x00 at %u): >%s<", index, input);

		LogEngine_SaveText("NMEAchecksum");

		return;
	}
	
	if (!strncmp( input, "$GPGGA", 6 )) {
		// continue processing the GGA entry only when a handler was provided
		if (Handler_GPS_NMEA_GGA != NULL) {
			// try to decode the entry
			logentry_t entry;
			if (NMEA_Read_GPGGA(input, &entry)) {
				// foward to the application for processing, e.g. logging (the check whether the handler was provided was already executed
				Handler_GPS_NMEA_GGA(input, &entry);				
			}	
		}
		// Every 10 fixes, trigger pressure sensor reading
        /*fix_counter++;
        if (fix_counter > 5) {
            SCP_TriggerMeasurement();
            fix_counter = 0;
        }*/
		return;
	}
	
	if (!strncmp( input, "$GPRMC", 6 )) {
		time_t time_gps;
		rawtime_t rawtime_gps;
		
		if (NMEA_Read_GPRMC(input, &time_gps, &rawtime_gps)) {
			
			if (RTC_ValidTime(&rawtime_gps)) {
				// the gps time is valid. compute the difference now.

				rawtime_t rawtime_system;
				RTC_GetRawTime(&rawtime_system);
				
				uint32_t timeDifference = 0;
				if (rawtime_system.seconds > rawtime_gps.seconds)
					timeDifference = rawtime_system.seconds - rawtime_gps.seconds;
				else
					timeDifference = rawtime_gps.seconds - rawtime_system.seconds;
				
				// if the difference is big enough, update the system time
				 
				if (timeDifference > 2) {
					// update
					RTC_SetRawTime(&rawtime_gps);

					// and log this event
					logentry_t entry;
					LogEngine_PrepareLogEntry(&entry);
					entry.TypeId = Debug;
					entry.Payload.Debug.DebugCode = 3; // updated the time
					entry.Payload.Debug.Value6_32bit = rawtime_system.seconds;
					entry.Payload.Debug.Value5_16bit = rawtime_system.miliseconds;
					LogEngine_Save(&entry);
				} // if the timediff was big enough
			} // if gps time is valid
			
			// received a valid RMC sentence from the device.
			// foward to the application for processing
			if (Handler_GPS_NMEA_RMC != NULL)
				Handler_GPS_NMEA_RMC(input, &time_gps, &rawtime_gps);
			
		} // if the RMC sentence was read
		
		return;
	}
}


void _GPS_NMEA_Debug_Handler(const char* input) {
	if (NMEA_ValidSentence(input) == false) {
		
		// try to find the end of string for debugging:
		uint8_t index;
		for (index = 0; index < 120; index++) {
			if (input[index] == 0x00)
				break;
		}

		LOGERR("CHKSUM FAIL (0x00 at %u): >%s<", index, input);
		
		return;
	}
	
	printf(input);
	printf("\r\n");
}

void GPS_PowerOn() {
	// turn on the power supply of the GPS chip
	SET(GPS_PWR_OUT, GPS_PWR_PIN);
	
	// wait for the device to start
	LOGDBG("GPS Powered Waiting 5 seconds");
	Timers_Block(TICKS_5SECONDS);

	LOGDBG("Powering Serial line");
	USCI_A1_PowerOn();
	
	LogEngine_SaveText("GPS_On");
}

void GPS_PowerOff() {
	// turn off the power supply of the GPS chip
	CLEAR(GPS_PWR_OUT, GPS_PWR_PIN);
	
	// dislable the interrupt on UART1 RX and unregister handlers
	USCI_A1_PowerOff();
	UART_A1_CharacterPreview = NULL;
	UART_A1_LineReceivedHandler = NULL;
	
	LogEngine_SaveText("GPS_Off");
}

void GPS_StartProcessingNMEA() {
	// test whether the GPS chip is enabled.
	// dont activate NMEA processing if GPS isn't enabled
	LOGDBG("Registering GPS NMEA Handler...");
	if (GPS_PWR_OUT & GPS_PWR_PIN) {
		UART_A1_CharacterPreview = NULL;
		UART_A1_LineReceivedHandler = _GPS_NMEA_Processing_Handler;
		LOGDBG("...Registered GPS NMEA Handler.");
	}
}

bool GPS_Init(){
	_command_replies_in_buffer = 0;
	
	LOGDBG("GPS Init... (up to 20 seconds)");
	
	// Initializing UART: no line handler (this one is important only while debugging)
	//USCI_UART_Init(USCI_A1, 9600L);
	USCI_UART_Init(USCI_A1, 4800L);
	UART_A1_CharacterPreview = NULL;
	UART_A1_LineReceivedHandler = NULL;
	
	SET(P3SEL, BIT6 | BIT7);

	// power on the GPS module
	SET(GPS_PWR_OUT, GPS_PWR_PIN);
	Timers_Block(TICKS_5SECONDS);

	// prepare for data communication with the GPS receiver
	UART_A1_CharacterPreview = _GPS_IncomingCommandResponseBytesHandler;
		

	/// 1. Make sure the GPS receiver communicates with 38400 baud
	
	// probe for the GPS device at 9600 baud and reconfigure if necessary
	GPS_SubmitCommand(gps_command_Test);
	bool ack_received;
	uint8_t buffer[64];	 // not used
	ack_received = GPS_BlockUntilPacketTypeReceived("001", buffer, sizeof(buffer));

	if (true == ack_received) {
		//LOGDBG("GPS: Detected at 9600 baud. Reconfiguring to 38400...");
		LOGDBG("GPS: Detected at 4800 baud. Reconfiguring to 38400...");
		
		GPS_SubmitCommand(gps_command_Set_NMEA_Baudrate_38400);
	} 
	
    ack_received = false;
	
	// Having trouble with timing here, let's try comminicating with the GPS a number of times...
	uint8_t attempts;
	for (attempts = 0; attempts < 10; attempts++) {
	    //avoid watchdog timer reset in the middle of the long run
		//... as it appears in rr=4 after 6 attempts
		Watchdog_Reset();

	    LOGDBG("GPS: Attempting comm at 38400...attempt %u", attempts);
	    GPS_SubmitCommand(gps_command_Set_NMEA_Baudrate_38400);
	    // Wait 2 seconds
	    Timers_Block(TICKS_2SECONDS);
	    
	    
	    // now, the GPS receiver should run at 38400 baud
	    // the UART of the MSP should be reconfigured as well
	    USCI_UART_Configure(USCI_A1, 38400L);
	    Timers_Block(TICKS_2SECONDS);
	    
	    // try to talk to the GPS again
	
	    GPS_SubmitCommand(gps_command_Test);
	    // Wait 2 seconds
	    Timers_Block(TICKS_2SECONDS);
	    ack_received = GPS_BlockUntilPacketTypeReceived("001", buffer, sizeof(buffer));
	    
		if (true == ack_received) {
            break;
	    } else {
	        // Fall back to 4800 and loop to try again
	        USCI_UART_Configure(USCI_A1, 4800L);
	    }
	}
	
	/*Timers_Block(TICKS_5SECONDS);
	Timers_Block(TICKS_5SECONDS);
	// try to talk to the GPS again
	
	GPS_SubmitCommand(gps_command_Test);
	Timers_Block(TICKS_5SECONDS);
	Timers_Block(TICKS_5SECONDS);
	ack_received = GPS_BlockUntilPacketTypeReceived("001", buffer, sizeof(buffer));
    */    
	if (true == ack_received) {
		LOGDBG("GPS: Detected at 38400 baud");
	} else {
		GPS_PowerOff();
		return false;
	}

    // Similarly, timing issues, try repeating NMEA configuration...
    
    for (attempts = 0; attempts < 10; attempts++) {
        LOGDBG("GPS: NMEA Configuration..attempt %u", attempts);

	    GPS_SubmitCommand(gps_command_Configure_NMEA_Messages);
	    //Timers_Block(TICKS_2SECONDS);
	    ack_received = GPS_BlockUntilPacketTypeReceived("001,314", buffer, sizeof(buffer));
        if (true == ack_received) {
            break;
	    } 
    }

	if (true == ack_received) {
		LOGDBG("GPS: NMEA configured: GGL: %c RMC: %c VTG %c GGA: %c GSA: %c GSV: %c GRS: %c GST: %c",
				gps_command_Configure_NMEA_Messages[4],
				gps_command_Configure_NMEA_Messages[6],
				gps_command_Configure_NMEA_Messages[8],
				gps_command_Configure_NMEA_Messages[10],
				gps_command_Configure_NMEA_Messages[12],
				gps_command_Configure_NMEA_Messages[14],
				gps_command_Configure_NMEA_Messages[16],
				gps_command_Configure_NMEA_Messages[18]
			);
	} else {
		LOGDBG("GPS: NMEA configuration not confirmed.");
		GPS_PowerOff();
		return false;
	}

	// stop monitoring the output of the GPS receiver
	
	/// 3. Set the position
	GPS_ChangePositionUpdateRate(Configuration.GpsRate);
	

	UART_A1_CharacterPreview = NULL;
	
	GPS_PowerOff();
	LOG("GPS Init: OK");
	
	return true;
}

bool GPS_ChangePositionUpdateRate(uint16_t requestedPositionUpdateRate) {
	if ((GPS_PWR_OUT & GPS_PWR_PIN) != GPS_PWR_PIN) {
		return false; // return if the GPS is not enabled.
	}

	UART_A1_CharacterPreview = _GPS_IncomingCommandResponseBytesHandler;

	uint16_t currentUpdateRate = 0;
	uint8_t buffer[64];	 // not used

	GPS_SubmitCommand(gps_command_Query_Position_Rate);
	if (GPS_BlockUntilPacketTypeReceived("500", buffer, sizeof(buffer))) {
		small_sscanf((char*) buffer, "500,%lu,", &currentUpdateRate);
	
		if (currentUpdateRate != requestedPositionUpdateRate) {
			char* command = NULL;
			switch (requestedPositionUpdateRate) {
				case 1:		command = (char*) gps_command_Position_Rate_1Hz; break;
				case 4:		command = (char*) gps_command_Position_Rate_4Hz; break;
				case 5:		command = (char*) gps_command_Position_Rate_5Hz; break;
				case 10:	command = (char*) gps_command_Position_Rate_10Hz; break;
				default:	break;
			}
			
			GPS_SubmitCommand(command);
			GPS_BlockUntilPacketTypeReceived("001,500", buffer, sizeof(buffer));
		}
	
	
		GPS_SubmitCommand(gps_command_Query_Position_Rate);
		if (GPS_BlockUntilPacketTypeReceived("500", buffer, sizeof(buffer))) {
			small_sscanf((char*) buffer, "500,%lu,", &currentUpdateRate);
			
			LOGDBG("GPS: Position Update Rate: %u Hz", 1000 / currentUpdateRate);
			UART_A1_CharacterPreview = NULL;
			return true;
		}
		
		UART_A1_CharacterPreview = NULL;
	}
	
	return false;
}

bool _GPS_IncomingBytesDebugHandler (uint8_t data) {
	putchar(data);
	LED_TOGGLE;

	SystemFlags.Flag.Pending_WakeUp_Request = true;
	return false;
}

COMMAND(gps, "GPS debug interface", cmd_args) {
	if (cmd_args->ArgumentCount < 1 || cmd_args->ArgumentCount > 2) {
		printf("Usage: gps [operation]\r\n"
				" 1:   Full NMEA debug mode (one way, reset to disable)\r\n"
				" 2:   Toggle NMEA preview mode\r\n"
				" 3:   Toggle Power of the GPS Chip\r\n"
				" 4:   Toggle NMEA Debug Handler\r\n"
				" 5:   Enable GPS, Start Logging (one way)\r\n"
				" 6 X: Set the position update rate to X (X = 1, 4, 5, or 10)\r\n"
				);
		return;
	}
	
	 switch (cmd_args->ArgumentArray[0][0]) {
		 case '1' : {
			 	GPS_PowerOn();

				UART_A1_CharacterPreview = _GPS_IncomingCommandResponseBytesHandler;
			 	
				GPS_SubmitCommand(gps_command_Configure_NMEA_Messages_All);

				GPS_SubmitCommand(gps_command_Position_Rate_10Hz);
			 	
				UART_A1_CharacterPreview = NULL;
				UART_A1_LineReceivedHandler = _GPS_NMEA_Debug_Handler;
				
				break;
		 }
		 case '2' : {
			 	if (UART_A1_CharacterPreview == _GPS_IncomingBytesDebugHandler)
			 		UART_A1_CharacterPreview = NULL;
			 	else
			 		UART_A1_CharacterPreview = _GPS_IncomingBytesDebugHandler;
			 	
				break;
		 }
		 case '3' : {
			 	if (GPS_PWR_OUT & GPS_PWR_PIN) {
				 	GPS_PowerOff();
					UART_A1_CharacterPreview = NULL;
					UART_A1_LineReceivedHandler = NULL;
				}
				else {
					GPS_PowerOn();
				}
			 	
			 	break;
		 }
		 case '4' : {
			 	if (UART_A1_LineReceivedHandler == _GPS_NMEA_Debug_Handler)
			 		UART_A1_LineReceivedHandler = NULL;
			 	else
			 		UART_A1_LineReceivedHandler = _GPS_NMEA_Debug_Handler;

				break;
		 }
		 
		 case '5' : {
			 	GPS_PowerOn();
			 	GPS_StartProcessingNMEA();
			 	
				break;
		 }
		 case '6' : {
				if (cmd_args->ArgumentCount == 2) {
					uint16_t newRate = atoi(cmd_args->ArgumentArray[1]);
					GPS_ChangePositionUpdateRate(newRate);
					
					Configuration.GpsRate = newRate;
					Configuration_Save();
				}
			 	break;
		 }
		 
		 default : {
			 // error
		 }
	 }
}




#ifndef DOXYGEN_PUBLIC_DOC
/**
* Copyright (c) Microsoft Corporation. All rights reserved.
* Licensed under the Apache License, Version 2.0 (the "License"); you may not use these files except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
* Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.
 
* Copyright (c) 2012 Freie Universitaet Berlin, Tomasz Naumowicz. All rights reserved.
* 
* Redistribution and use in source and binary forms, with or without modification, are
* permitted provided that the following conditions are met:
* 
* - Redistributions of source code must retain the above copyright notice, this list of
* conditions and the following disclaimer.
* 
* - Redistributions in binary form must reproduce the above copyright notice, this list
* of conditions and the following disclaimer in the documentation and/or other materials
* provided with the distribution.
* 
* - Neither the name of the Freie Universitaet Berlin nor the
* names of its contributors may be used to endorse or promote products
* derived from this software without specific prior written permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Freie Universitaet Berlin BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#endif // DOXYGEN
