/**
 * @file		
 * @brief		@b Driver: Flash Memory
 * For the S25FL128 and 256 P0XNFI001 (256KB sector module)
 *  
 *	 The memory is organized as:
 * 		- total 16,777,216 bytes (8 bits each)
 * 		- total 65536 pages (256 bytes each)
 * 		- total 256 sectors (256 pages each)
 * 
 * 	Each page can be individually programmed (bits are programmed from 1 to 0).
 * 
 * 	The device is Sector or Bulk Erasable (bits are erased from 0 to 1) but not Page Erasable.
 */

#ifndef S25FL128x256_API_H_
#define S25FL128x256_API_H_

#include <stdint.h>
#include "../System/Types.h"

/**
 * @brief			Write bytes to S25FL128S (or P)(or 256S) flash memory (assumes the memory was erased previously)
 * @param[in]		address			Address in the Flash Memory (000000 - FFFFFF)
 * @param[in]		buffer			Pointer to buffer with data to write
 * @param[in]		bytesToWrite	Number of bytes to write
 * 
 * It's not possible to overwrite data: bits are programmed from 1 to 0 and erased from 0 to 1
 */
 int16_t Flash_Write(uint32_t address, const void* buffer, uint16_t bytesToWrite);
 
/**
 * @brief			Read bytes from S25FL128S (or P)(or 256S) flash memory
 * @param[in]		address			Address in the Flash Memory (000000 - FFFFFF)
 * @param[out]		bufferOut		Pointer to buffer for read data
 * @param[in]		size			Number of bytes to read
 */
void Flash_Read(uint32_t address, void* bufferOut, uint16_t size);

/**
 * @brief	Erase entire S25FL128S (or P) flash memory
 */
void Flash_Erase();

/**
 * @brief			Blocks until the device to become available after writing or erasing.
 * 
 * @remarks			Use only when implementing handlers for very long running operations. All memory related functions already
 * 					have built-in support for waiting for the device implemented, but they are not watchdog safe and give no visual feedback
 * 					to the user.
 */
void Flash_WaitForDeviceWithVisualFeedback();

/**
 * @brief	Initializes the flash memory
 */
bool Flash_Init();

/**
 * @brief	Performs a self test of the flash memory
 */
bool Flash_SelfTest();

#endif /*S25FL128x256_API_H_*/

#ifndef DOXYGEN_PUBLIC_DOC
/**
* Copyright (c) Microsoft Corporation. All rights reserved.
* Licensed under the Apache License, Version 2.0 (the "License"); you may not use these files except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
* Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.
 
* Copyright (c) 2012 Freie Universitaet Berlin, Tomasz Naumowicz. All rights reserved.
* 
* Redistribution and use in source and binary forms, with or without modification, are
* permitted provided that the following conditions are met:
* 
* - Redistributions of source code must retain the above copyright notice, this list of
* conditions and the following disclaimer.
* 
* - Redistributions in binary form must reproduce the above copyright notice, this list
* of conditions and the following disclaimer in the documentation and/or other materials
* provided with the distribution.
* 
* - Neither the name of the Freie Universitaet Berlin nor the
* names of its contributors may be used to endorse or promote products
* derived from this software without specific prior written permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Freie Universitaet Berlin BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#endif // DOXYGEN
