/**
 * @file		
 * @brief		@b System: Management of the USCI (used mainly by drivers)
 */

#ifndef USCI_API_H_
#define USCI_API_H_

#include "Types.h"

/**
 * @name		Default configuration.
 * 
 * You can overwrite this configuration by putting the same defines with adjusted values in your \ref AppConfig.h
 * 
 * @{
 */

#ifndef UART0_RX_BUFFER_SIZE
	#define UART0_RX_BUFFER_SIZE					40 ///< Size of the receive buffer used of the UART0 (the serial debug output on the MATAKI) 
#endif

#ifndef UART1_RX_BUFFER_SIZE
	#define UART1_RX_BUFFER_SIZE					80  ///< Size of the receive buffer used of the UART1 (the GPS device output on the MATAKI) 
#endif

#ifndef UART_MAX_RECEIVED_LINE_LENGTH
	#define UART_MAX_RECEIVED_LINE_LENGTH			120 ///< Maximum length of a single string that can be received by the UART
#endif

#ifndef RESET_ON_ESC_ON_UART0
	#define RESET_ON_ESC_ON_UART0					1 ///< Controls the reset feature: when enabled a single ESC key can reset the device. Useful when debugging as no handler for a command needs to be executed in order to reset a device.
#endif

/** @} */

/**
 * @brief Selection of the USCI interface
 */
enum usci_id {
	USCI_A0		= 0,	///< USCI 0 (the serial debug output on the MATAKI borard, shared with SPI devices: radio, memory, SCP sensor)
	USCI_A1		= 1		///< USCI 1 (the GPS device on the MATAKI board)
};

/**
 * @brief @b Handler: Definition of the handler for processing of strings received via USCI
 * 
 * One parameter is supported:
 * 	- the received string
 */
typedef void (*fp_lineInputHandler_t) (const char* input);

/**
 * @ingroup	Handlers
 * @brief @b Handler: Access to the string (a single line) received via USCI A1 in UART mode (see \ref Handlers for more handlers).
 * 
 * You can access the process received strings in your application code by assigning 
 * your handler to this variable.
 * 
 * 
 * One parameter is supported:
 * 	- the received string
 * 
 * @remarks		Only one handler can be registered.
 * 				This handler is used by the \ref GPS.API.h GPS device. Use its handlers to access valid NMEA strings.
 */
fp_lineInputHandler_t UART_A1_LineReceivedHandler;

/**
 * @brief @b Handler: Definition of the handler for previewing characters received via USCI in UART mode
 * 
 * One parameter is supported:
 * 	- the received character
 * 
 * Returning true will cancel processing of this character by the System API.
 */
typedef bool (*fp_previewCharacter_t) (uint8_t);

/**
 * @ingroup	Handlers
 * @brief @b Handler: Previewing a character received via USCI A0 in UART mode (see \ref Handlers for more handlers).
 * 
 * You can preview and process characters received in your application code by assigning 
 * your handler to this variable.
 * 
 * One parameter is supported:
 * 	- the received character
 * 
 * Returning true stops further processing of this character by the System API.
 * 
 * @remarks		Only one handler can be registered.
 */
fp_previewCharacter_t UART_A0_CharacterPreview;

/**
 * @ingroup	Handlers
 * @brief @b Handler: Previewing a character received via USCI A1 in UART mode (see \ref Handlers for more handlers).
 * 
 * You can preview and process characters received in your application code by assigning 
 * your handler to this variable.
 * 
 * One parameter is supported:
 * 	- the received character
 * 
 * Returning true stops further processing of this character by the System API.
 * 
 * @remarks		Only one handler can be registered.
 * 				This handler is used by the \ref GPS.API.h GPS device.
 */
fp_previewCharacter_t UART_A1_CharacterPreview;


/**
 * @brief		Initializes the selected USCI
 * 
 * This function must be called first before the USCI module in UART mode can be used.
 * USCI A0 is initialized by the System API.
 * 
 * USCI 1 supports both: UART and SPI mode at the same time, no switching is required (comparted to the V4 version of the device)
 * 
 * Example:
 * @code
 * USCI_UART_Init(USCI_A0, 115200L);
 * @endcode
 * 
 * @remarks		In the MATAKI implementation, \ref GPS.API.h is initializing USCI A1.
 *
 * @param[in]	usci		The USCI to initialize
 * @param[in]	baudrate	The buadrate to be preconfigured.
 * 
 */
void USCI_UART_Init(const enum usci_id usci, const uint32_t baudrate);

/**
 * @brief		Reconfigures the selected UART
 * 
 * Updates the baudrate of the selected USCI interface in the UART mode.
 * 
 * @param[in]	usci		The USCI to initialize
 * @param[in]	baudrate	The buadrate to be preconfigured.
 * 
 */
void USCI_UART_Configure(enum usci_id usci, uint32_t baudrate);

/**
 * @name		SPI support for USCI A0
 * @{
 */

/**
 * @brief		Locks USCI B0 in the SPI mode with selected clock divicers
 * 
 * Calling this function locks the USCI B0 in the SPI mode. It's suggested to call this function
 * each time when starting to communicate with a sensor on the bus as this function gives feedback
 * whether the selected clock dividers are currently in use.
 * 
 * @remarks		Multlple locks are possible (they are counted).
 * 				The maximum MSP430F261x USCI bit clock frequency in SPI master mode is BRCLK, whereas on the
 * 				MSP430F16x USART module, it is BRCLK/2.
 * 
 * @param[in]	br0			BR0 divider
 * @param[in]	br1		`	BR1 divider
 * 
 * @returns		true on success
 */
bool 	USCI_B0_SPI_Lock(uint8_t br0, uint8_t br1);

/**
 * @brief		Releases the lock on USCI B0
 *
 * @remarks		Multiple locks are possible (they are counted). 
 */
void 	USCI_B0_SPI_Release();

/**
 * @brief		Transmits and receives a byte over the SPI on USCI B0
 * 
 * Transmit a 0x00 when only interested in receiving data.
 * 
 * @param[in]	outgoing_byte		Byte to transmit (transmit 0x00 when only interested in receiving)
 * 
 * @returns		Byte received.
 */
uint8_t USCI0_SPI_Comm(uint8_t outgoing_byte);

/** @} */

/**
 * @name		USCI A1 management
 * @{
 */

/**
 * @brief		Transmits a character over the USCI A1 in UART mode.
 * 
 * @param[in]	c Character to transmit
 * 
 * @remarks		The USCI A1 must be in the UART mode. This function blocks for 256 ACLK ticks before transmitting a character.
 * 
 * @returns		The transmitted character.
 */
int USCI_A1_UART_TX(int c);

/**
 * @brief		Transmits a buffer over the USCI A1 in UART mode.
 * 
 * @param[in]	length		The lenght of the buffer to transmit
 * @param[in]	buffer		The buffer to transmit
 * 
 * @remarks		The USCI A1 must be in the UART mode.
 */
void USCI_A1_UART_TX_Buffer(int length, const uint8_t* buffer);

/**
 * @brief		Enables the USCI A1. Defaults to the UART mode.
 */
void USCI_A1_PowerOn();

/**
 * @brief		Disables the USCI A1.
 */
void USCI_A1_PowerOff();

/** @} */

#endif /*USCI_API_H_*/


#ifndef DOXYGEN
/**
* Copyright (c) Microsoft Corporation. All rights reserved.
* Licensed under the Apache License, Version 2.0 (the "License"); you may not use these files except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
* Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.
 
* Copyright (c) 2012 Freie Universitaet Berlin, Tomasz Naumowicz. All rights reserved.
* 
* Redistribution and use in source and binary forms, with or without modification, are
* permitted provided that the following conditions are met:
* 
* - Redistributions of source code must retain the above copyright notice, this list of
* conditions and the following disclaimer.
* 
* - Redistributions in binary form must reproduce the above copyright notice, this list
* of conditions and the following disclaimer in the documentation and/or other materials
* provided with the distribution.
* 
* - Neither the name of the Freie Universitaet Berlin nor the
* names of its contributors may be used to endorse or promote products
* derived from this software without specific prior written permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Freie Universitaet Berlin BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#endif // DOXYGEN
