/**
 * @brief		Management of the USCI (used mainly by drivers)
 */


#include <msp430f2618.h>
#include <sys/cdefs.h>
#include <stdint.h>
#include <legacymsp430.h>
#include <mspgcc/ringbuffer.h>

#include <msp430/usart.h>

#include "BitOperations.h"
#include "Watchdog.h"
#include "Timers.h"
#include "System.h"
#include "System.API.h"
#include "Logging.h"

#include "USCI.h"

#define USCI_A0_RESET_RX()		{ IFG2 &= ~UCA0RXIFG; }
#define	USCI_A0_RESET_RXTX()		{ IFG2 &= ~(UCA0RXIFG | UCA0TXIFG); }

#define	USCI_A0_WAIT_RX()			while( (IFG2 & UCA0RXIFG) == 0 ) { _NOP(); }

#define USCI_A0_WAIT_TXDONE()		while ( !((IFG2 & UCA0TXIFG) == UCA0TXIFG) ) { _NOP(); }

#define USCI_A1_WAIT_TXDONE()		while( !((UC1IFG & UCA1TXIFG) == UCA1TXIFG) ) { _NOP(); }

/**
 * @brief	Describes USCI configurations, used for UART registers caching.
 */
typedef struct
{
	uint8_t br0;		///< Cached settings for the BR0 register
	uint8_t br1;		///< Cached settings for the BR1 register
	uint8_t mctl;		///< Cached settings for the MCTL register
} UART_Configuration_t;

/**
 * @brief	Stores USCI configurations, used for UART registers caching.
 */
static UART_Configuration_t _UART_Configurations[2];

//used to detect whether UART mode can be used
//it's a number so that the locks can be counted (imagine nested calls to spi interface with the same device/the same configuration)
static volatile uint8_t			_usci0_spi_locked;

//all characters received via UART are stored in this ringbuffer.
RINGBUFFER_NEW(_uart0_rx_ringbuffer, UART0_RX_BUFFER_SIZE);
RINGBUFFER_NEW(_uart1_rx_ringbuffer, UART1_RX_BUFFER_SIZE);

/**
 * it's important to keep track of the number of lines stored in the incoming ring buffer.
 * it's not enough to assume, there is a line already when there are bytes in the buffer
 * it's not safe to clear the new line flag after getting a line from the buffer when there could
 * be another one stored in the meantime.
 * 
 * soulution: count lines in the buffer, increase on add, decrease on remove. inconsistency could happen.
 * a better soulution might be implemented in future.
 */
static volatile uint8_t _uart1_rx_lines_in_buffer;

// this variable will be used to monitor the length of the ringbuffer for the uart1
uint16_t _uart1_rx_max_buffer_length;

/**
 * @brief	Computes configuration registers for selected usci and specified baudrate.
 * 
 * @author	Hans-Peter Heitzmann
 */ 
static void _PrepareUartConfiguration(enum usci_id usci, uint32_t baudrate) {
	unsigned long tmp_long = 0;
	unsigned int i;
	unsigned char mod, modmask;
	float frac, frac_sum;
	
	unsigned long system_clock = System_MeasureSMCLK();
	
	tmp_long = system_clock/baudrate;
	frac = (float)system_clock/(float)baudrate;
	frac -= tmp_long;						
	frac = 1.0 / frac;						
	frac_sum = frac;
	modmask = 0x80;
	mod = 0;
	
	if( frac > 0.1 )
	{
		for( i = 1; i < 9; i++ )
		{
			if( frac_sum < (float)(i + 1) && frac_sum > (float)(i - 1) )
			{
				mod |= modmask;
				modmask >>= 1;
				frac_sum += frac;
				continue;
			}
			modmask >>= 1;
		}
	}
	
	_UART_Configurations[usci].br0 = (unsigned char)( tmp_long & 0x00FF ); 
	_UART_Configurations[usci].br1 = (unsigned char)( tmp_long >> 8 );
	_UART_Configurations[usci].mctl = (unsigned char)mod;
}

/**
 * Interrupt handler for USCI A+B 0 in UART mode
 */
interrupt(USCIAB0RX_VECTOR) Usci0_RX_ISR(void) {
	/**
	 * this interrupt vector is shared between USCIA0 and USCIB0
	 */
	
	// continue only if it's generated by the USCIA
	if ((IFG2 & UCA0RXIFG) != UCA0RXIFG) {
		// cancel otherwise
		IFG2 &= ~UCB0RXIE;
		return;
	}
	
	/**
	 * no edge detection is required on the MSP430F2618 (compared to the previous MSP430F1611)
	 * the support for edge detection was removed
	 */

	uint8_t input;								// used to hold the received character

	input = UCA0RXBUF;							//read the rx buffer of the USCI
	
#if RESET_ON_ESC_ON_UART0

		if (input == 27) {						// 27 is the ASCII code for the ESC character
			__wd_resetinfo.Password = WD_RESETINFO_PWD;
			__wd_resetinfo.ResetReason = WD_RESET_REASON_REQUESTED;
			WDTCTL = 0xDEAD;
		}
		
#endif
	
	// allow to modify the behaviour of the ISR: the application can preview characters and interrupt its processing
	if (UART_A0_CharacterPreview != NULL) {		//if the preview handler was definded
		if (UART_A0_CharacterPreview(input))  	//use the preview handler: the handler returns TRUE if the processing of the characters should be stopped
			return;
	}
	
	// regular processing of entered character: storage in the ring buffer and notifications via system flags after the NEW LINE character
	
	if (input == '\n')	
		return;
	
	if (input == '\b')							//ingore backspace	
		return;
	
	if (input == '\r') {
		ringbuffer_put(&_uart0_rx_ringbuffer, '\0');
		SystemFlags.Flag.Pending_Uart0_NewLine = true;
		END_LPM_WITH_CONDITION_CHECK;		
		return;
	}

	ringbuffer_put(&_uart0_rx_ringbuffer, input);		//store all other characters
}

/**
 * Interrupt handler for USCI A+B 1 in UART mode
 */
interrupt(USCIAB1RX_VECTOR) Usci1_RX_ISR(void) {
	
	/**
	 * we should care only, if there is something that monitors UART in software anyway.
	 * so if there is no handler active, we can just ignore everything
	 * 
	 * this is done below:
	 */
	
	uint8_t input;								// used to hold the received character

	input = UCA1RXBUF;							//read the rx buffer of the USCI
	
	// allow to modify the behaviour of the ISR: the application can preview characters and interrupt its processing
	if (UART_A1_CharacterPreview != NULL) {		//if the preview handler was definded
		
		//use the preview handler: the handler returns TRUE if the processing of the characters should be stopped
		bool abort = UART_A1_CharacterPreview(input);

		// test the wake-up condition and wake up the uc if it was requested
		if (SystemFlags.Flag.Pending_WakeUp_Request) {
			SystemFlags.Flag.Pending_WakeUp_Request = false;
			END_LPM_WITH_CONDITION_CHECK;
		}
	
		// stop processing if the character was accquired by the character preview function
		if (abort)
			return;
	}
	
	// don't care, if nobody is registered
	if (UART_A1_LineReceivedHandler == NULL)
		return;
	
	// regular processing of entered character
	
	if (input == '\n')	
		return;
	
	if (input == '\r') {
		ringbuffer_put(&_uart1_rx_ringbuffer, '\0');
		_uart1_rx_lines_in_buffer++;
		SystemFlags.Flag.Pending_Uart1_NewLine = true;
		END_LPM_WITH_CONDITION_CHECK;		
		return;
	}

	int result = ringbuffer_put(&_uart1_rx_ringbuffer, input);		//store all other characters
	
	if (result == -1) {
		_uart1_rx_lines_in_buffer = 0;
		ringbuffer_clear(&_uart1_rx_ringbuffer);
	}
	
}


void USCIA0_UART_ReadLine(uint8_t* output) {
	int8_t index = -1;
	int8_t character;
	
	do {
		character = ringbuffer_get(&_uart0_rx_ringbuffer);
		if (character == -1)								// if the ringbuffer was empty: break the loop
			break;
		
		index++;
		output[index] = character;
	} while (output[index] != '\0' && index < UART_MAX_RECEIVED_LINE_LENGTH);
	
	/**
	 * the loop terminated: check if the data is correct
	 * the valid state is when:
	 * 		_lastReceivedLine[index] == '\0'
	 * it might be, that the received line was longer than the maximum allowed length
	 * in that case the loop breaks the execution and the '\0' would be missing.
	 * in that case the ring buffer could contain the rest of the message: the state of the ring buffer
	 * would be corrupted: it would still hold a message, that wouldn't be processed until the next NEW LINE
	 * is received.
	 * In order to make sure, the output is in the valid state, '\0' will be stored 
	 * at the end of the line.
	 */
	
	output[index] = '\0';
	
	/**
	 * in order to make sure that the ring buffer is in the correct state:
	 * if the character == -1: the buffer was empty, the buffer is in the correct state.
	 * the character != -1: there is still something in the buffer: set the fimrware flag to NEW LINE,
	 * so that the application is forced to pick up the data again.
	 * it might be, that the last character in the buffer was the '\0'
	 * this would mean, that the state of the ringbuffer is correct.
	 * in order to see, if data in the buffer is available, the length of the ringbuffer will be used:
	 */
	
	if (ringbuffer_len(&_uart0_rx_ringbuffer) > 0)
		SystemFlags.Flag.Pending_Uart0_NewLine = true;
}

/**
 * Reads a received string/line from the internal ring buffer
 */
void USCIA1_UART_ReadLine(uint8_t* output) {
	int8_t index = -1;
	int8_t character;
	
	uint16_t rbuff_len = ringbuffer_len(&_uart1_rx_ringbuffer);
	if (rbuff_len > _uart1_rx_max_buffer_length) {
		_uart1_rx_max_buffer_length = rbuff_len;
		LOG("UART1 RX BUFF LEN: %u", _uart1_rx_max_buffer_length);
	}
	
	do {
		character = ringbuffer_get(&_uart1_rx_ringbuffer);
		if (character == -1) {								// if the ringbuffer was empty: break the loop
			break;
		}
		
		index++;
		output[index] = character;
	} while (output[index] != '\0' && index < UART_MAX_RECEIVED_LINE_LENGTH);
	
	/**
	 * the loop terminated: check if the data is correct
	 * the valid state is when:
	 * 		_lastReceivedLine[index] == '\0'
	 * it might be, that the received line was longer than the maximum allowed length
	 * in that case the loop breaks the execution and the '\0' would be missing.
	 * in that case the ring buffer could contain the rest of the message: the state of the ring buffer
	 * would be corrupted: it would still hold a message, that wouldn't be processed until the next NEW LINE
	 * is received.
	 * In order to make sure, the output is in the valid state, '\0' will be stored 
	 * at the end of the line.
	 */
	
	output[index] = '\0';
	
	/**
	 * in order to make sure that the ring buffer is in the correct state:
	 * if the character == -1: the buffer was empty, the buffer is in the correct state.
	 * the character != -1: there is still something in the buffer: set the fimrware flag to NEW LINE,
	 * so that the application is forced to pick up the data again.
	 * it might be, that the last character in the buffer was the '\0'
	 * this would mean, that the state of the ringbuffer is correct.
	 * in order to see, if data in the buffer is available, the length of the ringbuffer will be used:
	 */
	
	/**
	 * there is an issue that can become visible in the following case:
	 * a line is put into the ringbuffer, but one character gets corrupted and is set to 0
	 * 	as a result, the ringbuffer contains "datadatadata0datadtadta0" but the lines_in_buffer counter is only set to 1
	 * then, data is read, and only the first part untio 0 is extracted. the rest is still in the buffer but the lines_in_buffer counter says 0...
	 * 
	 * when this happens, the most recent line will be lost and so this error can propagate...
	 * detection is difficult but -> once the buffer is full, the interrupt flushes all the data and executes internal uart1 logic reset
	 */ 

	if (_uart1_rx_lines_in_buffer > 0)
		_uart1_rx_lines_in_buffer--;

	if (_uart1_rx_lines_in_buffer > 0)
		SystemFlags.Flag.Pending_Uart1_NewLine = true;
}

/**
 * Implementation of the putchar function from the stdio.h
 * This function is required for printf etc to work
 */
int putchar(int c) {
	USCI_A0_WAIT_TXDONE();
	UCA0TXBUF  = c;
	USCI_A0_WAIT_TXDONE();
	
	return c;
}

/**
 * @brief	This function has a delay built in as it was developed parallel to work with the GPS chip attached to the USCI A1 on the MATAKI V4. It should be updated in the future
 */
int USCI_A1_UART_TX(int c) {
	Timers_Block(64);
	
	USCI_A1_WAIT_TXDONE();
	UCA1TXBUF = c;
	
	return c;
}

void USCI_A1_UART_TX_Buffer(int length, const uint8_t* buffer) {
	uint8_t index;
	for (index = 0; index < length; index++) {
		USCI_A1_UART_TX(buffer[index]);
	}
}


void _USCI_A0_UART_Reconfigure() {
	USCI_A0_WAIT_TXDONE();
	
	//this is a critical section, we have to disable all interrupts.
	_DINT();
	
	//	The recommended USCI initialization/re-configuration process is:
	//	1. Set UCSWRST (BIS.B #UCSWRST,&UCAxCTL1)
	UCA0CTL1 |= UCSWRST;	
	
	//	2. Initialize all USCI registers with UCSWRST = 1 (including UCAxCTL1)
	UCA0CTL1 |= UCSSEL_3;

	//set the speed according to the stored configuration
	UCA0BR0 = _UART_Configurations[USCI_A0].br0;
	UCA0BR1 = _UART_Configurations[USCI_A0].br1;
	UCA0MCTL = _UART_Configurations[USCI_A0].mctl;
			
	//	3. Configure ports. done already
	
	//	4. Clear UCSWRST via software (BIC.B #UCSWRST,&UCAxCTL1)
	UCA0CTL1 &= ~UCSWRST;

	//	5. Enable interrupts (optional) via UCAxRXIE and/or UCAxTXIE	
	IE2 |= UCA0RXIE; //interrupt enable for RX
	
	//re-enable interrupts
	_EINT();
}

void _USCI_A1_UART_Reconfigure() {
	USCI_A1_WAIT_TXDONE();
	
	//this is a critical section, we have to disable all interrupts.
	_DINT();
	
	//	The recommended USCI initialization/re-configuration process is:
	//	1. Set UCSWRST (BIS.B #UCSWRST,&UCAxCTL1)
	UCA1CTL1 |= UCSWRST;	
	
	//	2. Initialize all USCI registers with UCSWRST = 1 (including UCAxCTL1)
	UCA1CTL1 |= UCSSEL_3;

	//set the speed according to the stored configuration
	UCA1BR0 = _UART_Configurations[USCI_A1].br0;
	UCA1BR1 = _UART_Configurations[USCI_A1].br1;
	UCA1MCTL = _UART_Configurations[USCI_A1].mctl;
			
	//	3. Configure ports.
	SET(P3SEL, BIT6 | BIT7);
	
	//	4. Clear UCSWRST via software (BIC.B #UCSWRST,&UCAxCTL1)
	UCA1CTL1 &= ~UCSWRST;

	//	5. Enable interrupts (optional) via UCAxRXIE and/or UCAxTXIE	
	UC1IE |= UCA1RXIE; //interrupt enable for RX
	
	//re-enable interrupts
	_EINT();
}

void USCI_A1_PowerOn() {
	// flush the buffer first
	_uart1_rx_lines_in_buffer = 0;
	_uart1_rx_max_buffer_length = 0;
	
	ringbuffer_clear(&_uart1_rx_ringbuffer);
	
	SET(UC1IE, UCA1RXIE);		//interrupt enable for RX
	SET(P3SEL, BIT6 | BIT7);	//enable 2nd fct mode of the UTX1 and URX1
}

void USCI_A1_PowerOff() {
	CLEAR(UC1IE, UCA1RXIE); 	//disable interrupts on RX1
	CLEAR(P3SEL, BIT6 | BIT7);	//disable 2nd fct mode of the UTX1 and URX1, this fallbacks to GPIO mode where both pins are configured as output with value 0 (important so that the GPS doesn't steal power)	
}

void USCI_UART_Init(const enum usci_id usci, const uint32_t baudrate) {
	switch (usci) {
		case USCI_A0: {
			_usci0_spi_locked = 0;
			UART_A0_CharacterPreview = NULL;
			
			USCI_UART_Configure(USCI_A0, baudrate);
			
			break;
		}
		case USCI_A1: {
			_uart1_rx_lines_in_buffer = 0;

			USCI_UART_Configure(USCI_A1, baudrate);
			
			break;
		}
	}
}

void USCI_UART_Configure(enum usci_id usci, uint32_t baudrate) {
	//prepare parameters for the current frequency (the output is cached by the called function)
	_PrepareUartConfiguration(usci, baudrate);
	
	//reconfigure the system if not locked. currently USCI A1 is used only for GPS, so we don't care about SPI implementation for USCI 1 yet
	switch (usci) {
		case USCI_A0: {
			_USCI_A0_UART_Reconfigure();
		}
		case USCI_A1: {
			_USCI_A1_UART_Reconfigure();			
		}
	}
}

//mctl is not required for SPI
bool USCI_B0_SPI_Lock(uint8_t br0, uint8_t br1) {
	if (_usci0_spi_locked) {
		//the SPI is already locked in a specific configuration
		//if the new configuration is different, no lock can be accquired
		if ((UCB0BR0 == br0 && UCB0BR1 == br1)) {
			_usci0_spi_locked++;
			return true; // no reconfiguration is required
		} else {
			return false; // configurations differ, no lock can be accquired
		}
	}

	//The recommended USCI initialization/re-configuration process is:
	//1. Set UCSWRST (BIS.B #UCSWRST,&UCxCTL1)
	UCB0CTL1 = UCSWRST;
	
	//2. Initialize all USCI registers with UCSWRST=1 (including UCxCTL1)
	UCB0CTL0 = UCMST | UCSYNC | UCMSB | UCCKPH; // SPI master, Synchronous mode
	UCB0CTL1 |= UCSSEL_3;
	
	//set the speed according to the stored configuration
	UCB0BR0 = br0;
	UCB0BR1 = br1;
	
	//3. Configure ports. already done.

	//4. Clear UCSWRST via software (BIC.B #UCSWRST,&UCxCTL1)
	UCB0CTL1 &= ~UCSWRST;	

	// 5) Enable/Disable interrupts (optional) via the IEx SFRs (URXIEx and/or UTXIEx)

	SET(P3SEL, BIT1 | BIT2 | BIT3);		//make sure SPI0 pins are the 2nd fct mode
	
	_usci0_spi_locked++;
	
	return true;
}

void USCI_B0_SPI_Release() {
	if (_usci0_spi_locked == 0) {
		CLEAR(P3SEL, BIT1 | BIT2 | BIT3);		//make sure SPI0 pins are not in the 2nd fct mode
		
		return;
	}
	
	if (_usci0_spi_locked)
		_usci0_spi_locked--;
}


uint8_t USCI0_SPI_Comm(uint8_t outgoing_byte) {
	//return 0 if not in SPI mode
	if (!_usci0_spi_locked)
		return 0;

	uint8_t received_byte = UCB0RXBUF;		//dummy read
	
	UCB0TXBUF = outgoing_byte;		// Send data
	//TIMERS_BLOCK_MS(10);
	while((IFG2 & UCB0RXIFG) == 0) { _NOP(); };	//UCxRXIFG, is set, indicating the RX/TX operation is complete. Wait for RX (TX) to finish (could sleep now)

	received_byte = UCB0RXBUF;
	
	//LOGDBG("out: %u / in: %u", outgoing_byte, received_byte);
	
	return received_byte;			// Read data    
}

#ifndef DOXYGEN
/**
* Copyright (c) Microsoft Corporation. All rights reserved.
* Licensed under the Apache License, Version 2.0 (the "License"); you may not use these files except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
* Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.
 
* Copyright (c) 2012 Freie Universitaet Berlin, Tomasz Naumowicz. All rights reserved.
* 
* Redistribution and use in source and binary forms, with or without modification, are
* permitted provided that the following conditions are met:
* 
* - Redistributions of source code must retain the above copyright notice, this list of
* conditions and the following disclaimer.
* 
* - Redistributions in binary form must reproduce the above copyright notice, this list
* of conditions and the following disclaimer in the documentation and/or other materials
* provided with the distribution.
* 
* - Neither the name of the Freie Universitaet Berlin nor the
* names of its contributors may be used to endorse or promote products
* derived from this software without specific prior written permission.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Freie Universitaet Berlin BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#endif // DOXYGEN
